# -*- coding: utf-8 -*-

"""Operator plugin base class that all __subclassess__ are brought in scope."""
from __future__ import unicode_literals

from abc import ABCMeta, abstractmethod
from os import listdir as _listdir  # To not conflict with operator
from os.path import dirname, basename

# TODO: Allow for imports of custom operators and subdirectories.
__all__ = [
    basename(f)[:-3]
    for f in _listdir(dirname(__file__))
    if f[-3:] == ".py" and not f.endswith("__init__.py")
]


class BaseOperator(metaclass=ABCMeta):
    """Base operator mixin class."""

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize BaseOperator."""
        self.operator_dict = operator_dict
        self.context = context or {}
        self.no_input = no_input
        self.post_gen_operator = False

    @abstractmethod
    def execute(self):
        """Run the operator."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-

"""Operator plugin that inherits a base class and is made available through `type`."""
from __future__ import unicode_literals
from __future__ import print_function

import cookiecutter as cc
import logging

from cookiecutter.operators import BaseOperator


logger = logging.getLogger(__name__)


class NukikataOperator(BaseOperator):
    """Operator for calling external cookiecutters."""

    type = 'nukikata'

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize operator."""
        super(NukikataOperator, self).__init__(
            operator_dict=operator_dict, context=context, no_input=no_input
        )
        # Doesn't work, gets called twice
        self.post_gen_operator = (
            self.operator_dict['delay'] if 'delay' in self.operator_dict else False
        )

    def execute(self):
        """Run the nukikata operator."""
        return cc.main.cookiecutter(
            template=self.operator_dict['template'],
            checkout=self.operator_dict['checkout']
            if 'checkout' in self.operator_dict
            else None,
            no_input=self.no_input,
            context_file=self.operator_dict['context_file']
            if 'context_file' in self.operator_dict
            else None,
            extra_context=self.operator_dict['extra_context']
            if 'extra_context' in self.operator_dict
            else None,
            replay=self.operator_dict['replay']
            if 'replay' in self.operator_dict
            else None,
            overwrite_if_exists=self.operator_dict['overwrite_if_exists']
            if 'overwrite_if_exists' in self.operator_dict
            else False,
            output_dir=self.operator_dict['output_dir']
            if 'output_dir' in self.operator_dict
            else '.',
            config_file=self.operator_dict['config_file']
            if 'config_file' in self.operator_dict
            else None,
            default_config=self.operator_dict['default_config']
            if 'default_config' in self.operator_dict
            else False,
            password=self.operator_dict['password']
            if 'password' in self.operator_dict
            else None,
            directory=self.operator_dict['directory']
            if 'directory' in self.operator_dict
            else None,
            skip_if_file_exists=self.operator_dict['skip_if_file_exists']
            if 'skip_if_file_exists' in self.operator_dict
            else False,
        )


# # TODO: Is this needed?  Once we fix the output of the
# #  normal cookiecutter this won't be needed likely
class NukikataPromptOperator(BaseOperator):
    """Operator for nukikata type prompts."""

    type = 'nukikata_prompt'

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize operator."""
        super(NukikataPromptOperator, self).__init__(
            operator_dict=operator_dict, context=context, no_input=no_input
        )

    def execute(self):
        """Run the operator."""
        context = {'cookiecutter': {}}
        if 'context_file' in self.operator_dict:
            context['cookiecutter'] = cc.utils.read_config_file(
                self.operator_dict['template']
            )
        else:
            context['cookiecutter'] = self.operator_dict['context']

        return cc.prompt.prompt_for_config(
            context=context,
            no_input=self.operator_dict['no_input']
            if 'no_input' in self.operator_dict
            else False,
            context_key=self.operator_dict['context_key']
            if 'context_key' in self.operator_dict
            else None,
        )



"""Utility functions for handling and fetching repo archives in zip format."""

from __future__ import absolute_import

import os
import tempfile
from zipfile import ZipFile

import requests

try:
    # BadZipfile was renamed to BadZipFile in Python 3.2.
    from zipfile import BadZipFile
except ImportError:
    from zipfile import BadZipfile as BadZipFile

from cookiecutter.exceptions import InvalidZipRepository
from cookiecutter.prompt import read_repo_password
from cookiecutter.utils import make_sure_path_exists, prompt_and_delete


def unzip(zip_uri, is_url, clone_to_dir='.', no_input=False, password=None):
    """Download and unpack a zipfile at a given URI.

    This will download the zipfile to the cookiecutter repository,
    and unpack into a temporary directory.

    :param zip_uri: The URI for the zipfile.
    :param is_url: Is the zip URI a URL or a file?
    :param clone_to_dir: The cookiecutter repository directory
        to put the archive into.
    :param no_input: Suppress any prompts
    :param password: The password to use when unpacking the repository.
    """
    # Ensure that clone_to_dir exists
    clone_to_dir = os.path.expanduser(clone_to_dir)
    make_sure_path_exists(clone_to_dir)

    if is_url:
        # Build the name of the cached zipfile,
        # and prompt to delete if it already exists.
        identifier = zip_uri.rsplit('/', 1)[1]
        zip_path = os.path.join(clone_to_dir, identifier)

        if os.path.exists(zip_path):
            download = prompt_and_delete(zip_path, no_input=no_input)
        else:
            download = True

        if download:
            # (Re) download the zipfile
            r = requests.get(zip_uri, stream=True)
            with open(zip_path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
    else:
        # Just use the local zipfile as-is.
        zip_path = os.path.abspath(zip_uri)

    # Now unpack the repository. The zipfile will be unpacked
    # into a temporary directory
    try:
        zip_file = ZipFile(zip_path)

        if len(zip_file.namelist()) == 0:
            raise InvalidZipRepository('Zip repository {} is empty'.format(zip_uri))

        # The first record in the zipfile should be the directory entry for
        # the archive. If it isn't a directory, there's a problem.
        first_filename = zip_file.namelist()[0]
        if not first_filename.endswith('/'):
            raise InvalidZipRepository(
                'Zip repository {} does not include '
                'a top-level directory'.format(zip_uri)
            )

        # Construct the final target directory
        project_name = first_filename[:-1]
        unzip_base = tempfile.mkdtemp()
        unzip_path = os.path.join(unzip_base, project_name)

        # Extract the zip file into the temporary directory
        try:
            zip_file.extractall(path=unzip_base)
        except RuntimeError:
            # File is password protected; try to get a password from the
            # environment; if that doesn't work, ask the user.
            if password is not None:
                try:
                    zip_file.extractall(path=unzip_base, pwd=password.encode('utf-8'))
                except RuntimeError:
                    raise InvalidZipRepository(
                        'Invalid password provided for protected repository'
                    )
            elif no_input:
                raise InvalidZipRepository(
                    'Unable to unlock password protected repository'
                )
            else:
                retry = 0
                while retry is not None:
                    try:
                        password = read_repo_password('Repo password')
                        zip_file.extractall(
                            path=unzip_base, pwd=password.encode('utf-8')
                        )
                        retry = None
                    except RuntimeError:
                        retry += 1
                        if retry == 3:
                            raise InvalidZipRepository(
                                'Invalid password provided ' 'for protected repository'
                            )

    except BadZipFile:
        raise InvalidZipRepository(
            'Zip repository {} is not a valid zip archive:'.format(zip_uri)
        )

    return unzip_path



# -*- coding: utf-8 -*-

"""Tests to verify correct work with user configs and system/user variables inside."""

import os
import shutil

import pytest

from cookiecutter import config
from cookiecutter.exceptions import InvalidConfiguration


@pytest.fixture(scope='module')
def user_config_path():
    """Fixture. Return user config path for current user."""
    return os.path.expanduser('~/.cookiecutterrc')


@pytest.fixture(scope='function')
def back_up_rc(user_config_path):
    """
    Back up an existing cookiecutter rc and restore it after the test.

    If ~/.cookiecutterrc is pre-existing, move it to a temp location
    """
    user_config_path_backup = os.path.expanduser('~/.cookiecutterrc.backup')

    if os.path.exists(user_config_path):
        shutil.copy(user_config_path, user_config_path_backup)
        os.remove(user_config_path)

    yield
    # Remove the ~/.cookiecutterrc that has been created in the test.
    if os.path.exists(user_config_path):
        os.remove(user_config_path)

    # If it existed, restore the original ~/.cookiecutterrc.
    if os.path.exists(user_config_path_backup):
        shutil.copy(user_config_path_backup, user_config_path)
        os.remove(user_config_path_backup)


@pytest.fixture
def custom_config():
    """Fixture. Return expected custom configuration for future tests validation."""
    return {
        'default_context': {
            'full_name': 'Firstname Lastname',
            'email': 'firstname.lastname@gmail.com',
            'github_username': 'example',
        },
        'cookiecutters_dir': '/home/example/some-path-to-templates',
        'replay_dir': '/home/example/some-path-to-replay-files',
        'abbreviations': {
            'gh': 'https://github.com/{0}.git',
            'gl': 'https://gitlab.com/{0}.git',
            'bb': 'https://bitbucket.org/{0}',
            'helloworld': 'https://github.com/hackebrot/helloworld',
        },
    }


@pytest.mark.usefixtures('back_up_rc')
def test_get_user_config_valid(user_config_path, custom_config):
    """Validate user config correctly parsed if exist and correctly formatted."""
    shutil.copy('tests/test-config/valid-config.yaml', user_config_path)
    conf = config.get_user_config()

    assert conf == custom_config


@pytest.mark.usefixtures('back_up_rc')
def test_get_user_config_invalid(user_config_path):
    """Validate `InvalidConfiguration` raised when provided user config malformed."""
    shutil.copy('tests/test-config/invalid-config.yaml', user_config_path)
    with pytest.raises(InvalidConfiguration):
        config.get_user_config()


@pytest.mark.usefixtures('back_up_rc')
def test_get_user_config_nonexistent():
    """Validate default app config returned, if user does not have own config."""
    assert config.get_user_config() == config.DEFAULT_CONFIG


@pytest.fixture
def custom_config_path():
    """Fixture. Return path to custom user config for tests."""
    return 'tests/test-config/valid-config.yaml'


def test_specify_config_path(mocker, custom_config_path, custom_config):
    """Validate provided custom config path should be respected and parsed."""
    spy_get_config = mocker.spy(config, 'get_config')

    user_config = config.get_user_config(custom_config_path)
    spy_get_config.assert_called_once_with(custom_config_path)

    assert user_config == custom_config


def test_default_config_path(user_config_path):
    """Validate app configuration. User config path should match default path."""
    assert config.USER_CONFIG_PATH == user_config_path


def test_default_config_from_env_variable(
    monkeypatch, custom_config_path, custom_config
):
    """Validate app configuration. User config path should be parsed from sys env."""
    monkeypatch.setenv('COOKIECUTTER_CONFIG', custom_config_path)

    user_config = config.get_user_config()
    assert user_config == custom_config


def test_force_default_config(mocker, custom_config_path):
    """Validate `default_config=True` should ignore provided custom user config."""
    spy_get_config = mocker.spy(config, 'get_config')

    user_config = config.get_user_config(custom_config_path, default_config=True)

    assert user_config == config.DEFAULT_CONFIG
    assert not spy_get_config.called


def test_expand_user_for_directories_in_config(monkeypatch):
    """Validate user pointers expanded in user configs."""

    def _expanduser(path):
        return path.replace('~', 'Users/bob')

    monkeypatch.setattr('os.path.expanduser', _expanduser)

    config_file = 'tests/test-config/config-expand-user.yaml'

    user_config = config.get_user_config(config_file)
    assert user_config['replay_dir'] == 'Users/bob/replay-files'
    assert user_config['cookiecutters_dir'] == 'Users/bob/templates'


def test_expand_vars_for_directories_in_config(monkeypatch):
    """Validate environment variables expanded in user configs."""
    monkeypatch.setenv('COOKIES', 'Users/bob/cookies')

    config_file = 'tests/test-config/config-expand-vars.yaml'

    user_config = config.get_user_config(config_file)
    assert user_config['replay_dir'] == 'Users/bob/cookies/replay-files'
    assert user_config['cookiecutters_dir'] == 'Users/bob/cookies/templates'



# -*- coding: utf-8 -*-

"""Functions for discovering and executing various cookiecutter operators."""
from cookiecutter.environment import StrictEnvironment, render_variable
from cookiecutter.operators import *  # noqa

from cookiecutter.operators import BaseOperator

post_gen_operator_list = []


def run_operator(operator_dict: dict, context=None, no_input=False):
    """Run operator."""
    if context is None:
        context = {}
    operator_output = None
    delayed_output = None
    operator_list = BaseOperator.__subclasses__()
    for o in operator_list:
        if operator_dict['type'] == o.type:  # noqa
            operator = o(operator_dict, context, no_input)
            # TODO: Determine if default method is needed across all operators
            # If so, need to feed in `no_input` from both cases in call
            if operator.post_gen_operator:
                delayed_output = operator
            else:
                operator_output = operator.execute()
            break

    return operator_output, delayed_output


def parse_operator(
    context,
    key,
    cookiecutter_dict,
    append_key: bool = False,
    no_input: bool = False,
    context_key=None,
):
    """Parse input dict for loop and when logic and calls hooks.

    :return: cookiecutter_dict # noqa
    """
    if not context_key:
        context_key = next(iter(context))

    env = StrictEnvironment(context=context)
    operator_dict = context[context_key][key]

    global post_gen_operator_list

    if 'when' in operator_dict:
        if not context:
            raise ValueError("Can't have when condition without establishing context")

        when_condition = (
            render_variable(env, operator_dict['when'], cookiecutter_dict, context_key)
            == 'True'
        )

        operator_dict.pop('when')
        if not isinstance(when_condition, bool):
            raise ValueError("When condition needs to render with jinja to boolean")

    else:
        when_condition = True

    if when_condition:
        # Extract loop
        if 'loop' in operator_dict:
            loop_targets = render_variable(
                env, operator_dict['loop'], cookiecutter_dict, context_key
            )
            operator_dict.pop('loop')

            loop_output = []
            for l in loop_targets:
                loop_cookiecutter = cookiecutter_dict
                loop_cookiecutter.update({'item': l})
                loop_output += [
                    parse_operator(
                        context,
                        key,
                        loop_cookiecutter,
                        append_key=True,
                        no_input=no_input,
                    )
                ]

            cookiecutter_dict.pop('item')
            cookiecutter_dict[key] = loop_output
            return cookiecutter_dict

        operator_dict = render_variable(
            env, operator_dict, cookiecutter_dict, context_key
        )
        cookiecutter_dict[key], post_gen_operator = run_operator(
            operator_dict, context, no_input
        )  # output is list
        if post_gen_operator:
            post_gen_operator_list.append(post_gen_operator)

        if append_key:
            return cookiecutter_dict[key]

    return cookiecutter_dict



# -*- coding: utf-8 -*-

"""Operator plugin that inherits a base class and is made available through `type`."""
from __future__ import unicode_literals
from __future__ import print_function

from jinja2 import FileSystemLoader
import logging

from cookiecutter.operators import BaseOperator
from cookiecutter.environment import StrictEnvironment

logger = logging.getLogger(__name__)


class JinjaOperator(BaseOperator):
    """Operator for jinja templates."""

    type = 'jinja'

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize operator."""  # noqa
        super(JinjaOperator, self).__init__(
            operator_dict=operator_dict, context=context, no_input=no_input
        )
        # Defaulting to run inline
        self.post_gen_operator = (
            self.operator_dict['delay'] if 'delay' in self.operator_dict else False
        )
        self.file_system_loader = (
            self.operator_dict['file_system_loader']
            if 'file_system_loader' in self.operator_dict
            else '.'
        )

    def execute(self):
        """Run the operator."""  # noqa
        env = StrictEnvironment(context=self.context)

        env.loader = FileSystemLoader(self.file_system_loader)
        template = env.get_template(self.operator_dict['template_path'])

        output_from_parsed_template = template.render(**self.context)
        with open(self.operator_dict['output_path'], 'w') as fh:
            fh.write(output_from_parsed_template)

        return self.operator_dict['output_path']



# -*- coding: utf-8 -*-

"""Operator plugin that inherits a base class and is made available through `type`."""
from __future__ import unicode_literals
from __future__ import print_function

import logging
from PyInquirer import prompt

from cookiecutter.operators import BaseOperator

logger = logging.getLogger(__name__)


class InquirerPasswordOperator(BaseOperator):
    """Operator for PyInquirer `password` type prompts."""

    type = 'password'

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize PyInquirer operator."""  # noqa
        super(InquirerPasswordOperator, self).__init__(
            operator_dict=operator_dict, context=context, no_input=no_input
        )

    def execute(self):
        """Run the prompt."""  # noqa
        if 'name' not in self.operator_dict:
            self.operator_dict.update({'name': 'tmp'})
            return prompt([self.operator_dict])['tmp']
        else:
            return prompt([self.operator_dict])



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.prompt` module."""
from cookiecutter.operators.stat import StatOperator

FIXTURES = ['checkbox_bad.yaml']

context = {
    'cookiecutter': {
        'project_name': "Slartibartfast",
        'stuff': ['things', 'bears'],
        'foo': {'type': 'stat', 'input': {'bar': 'baz', 'grrr': 'wowzaz'}},
    }
}


def test_stat():
    """Verify simplest functionality."""
    c = StatOperator(context['cookiecutter']['foo'], context)
    d = c.execute()
    assert d == {'bar': 'baz', 'grrr': 'wowzaz'}



# -*- coding: utf-8 -*-

"""Helper functions used throughout Cookiecutter."""

from __future__ import unicode_literals
import contextlib
import errno
import logging
import os
import stat
import shutil
import sys
import json
import yaml
from _collections import OrderedDict

from cookiecutter.prompt import read_user_yes_no

logger = logging.getLogger(__name__)


def force_delete(func, path, exc_info):
    """Error handler for `shutil.rmtree()` equivalent to `rm -rf`.

    Usage: `shutil.rmtree(path, onerror=force_delete)`
    From stackoverflow.com/questions/1889597
    """
    os.chmod(path, stat.S_IWRITE)
    func(path)


def rmtree(path):
    """Remove a directory and all its contents. Like rm -rf on Unix.

    :param path: A directory path.
    """
    shutil.rmtree(path, onerror=force_delete)


def make_sure_path_exists(path):
    """Ensure that a directory exists.

    :param path: A directory path.
    """
    logger.debug('Making sure path exists: %s', path)
    try:
        os.makedirs(path)
        logger.debug('Created directory at: %s', path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            return False
    return True


@contextlib.contextmanager
def work_in(dirname=None):
    """Context manager version of os.chdir.

    When exited, returns to the working directory prior to entering.
    """
    curdir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)


def make_executable(script_path):
    """Make `script_path` executable.

    :param script_path: The file to change
    """
    status = os.stat(script_path)
    os.chmod(script_path, status.st_mode | stat.S_IEXEC)


def prompt_and_delete(path, no_input=False):
    """
    Ask user if it's okay to delete the previously-downloaded file/directory.

    If yes, delete it. If no, checks to see if the old version should be
    reused. If yes, it's reused; otherwise, Cookiecutter exits.

    :param path: Previously downloaded zipfile.
    :param no_input: Suppress prompt to delete repo and just delete it.
    :return: True if the content was deleted
    """
    # Suppress prompt if called via API
    if no_input:
        ok_to_delete = True
    else:
        question = (
            "You've downloaded {} before. " "Is it okay to delete and re-download it?"
        ).format(path)

        ok_to_delete = read_user_yes_no(question, 'yes')

    if ok_to_delete:
        if os.path.isdir(path):
            rmtree(path)
        else:
            os.remove(path)
        return True
    else:
        ok_to_reuse = read_user_yes_no(
            "Do you want to re-use the existing version?", 'yes'
        )

        if ok_to_reuse:
            return False

        sys.exit()


def read_config_file(file):
    """Read files into objects."""
    file_extension = file.split('.')[-1]

    if not os.path.exists(file):
        raise FileNotFoundError

    logger.debug(
        'Using \"{}\" as input file and \"{}\" as file extension'.format(
            file, file_extension
        )
    )
    if file_extension == 'json':
        with open(file) as f:
            config = json.load(f, object_pairs_hook=OrderedDict)
        return config
    elif file_extension in ('yaml', 'yml', 'nukirc'):
        with open(file, encoding='utf-8') as f:
            config = yaml.load(f, Loader=yaml.FullLoader)
        return config
    # elif file_extension == 'hcl':
    #     with open(file) as f:
    #         config = hcl.loads(f.read())
    #     return config
    else:
        raise ValueError(
            'Unable to parse file {}. Error: Unsupported extension (json/yaml only)'
            ''.format(file)
        )  # noqa



# -*- coding: utf-8 -*-

"""Test work of python and shell hooks for generated projects."""

from __future__ import unicode_literals

import errno
import os
import sys

import pytest

from cookiecutter import generate, utils
from cookiecutter.exceptions import FailedHookException

WINDOWS = sys.platform.startswith('win')


@pytest.fixture(scope='function')
def remove_additional_folders(tmpdir):
    """Remove some special folders which are created by the tests."""
    yield
    directories_to_delete = [
        'tests/test-pyhooks/inputpyhooks',
        'inputpyhooks',
        'inputhooks',
        os.path.join(str(tmpdir), 'test-shellhooks'),
        'tests/test-hooks',
    ]
    for directory in directories_to_delete:
        if os.path.exists(directory):
            utils.rmtree(directory)


@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_ignore_hooks_dirs():
    """Verify hooks directory not created in target location on files generation."""
    generate.generate_files(
        context={'cookiecutter': {'pyhooks': 'pyhooks'}},
        repo_dir='tests/test-pyhooks/',
        output_dir='tests/test-pyhooks/',
    )
    assert not os.path.exists('tests/test-pyhooks/inputpyhooks/hooks')


@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_run_python_hooks():
    """Verify pre and post generation python hooks executed and result in output_dir.

    Each hook should create in target directory. Test verifies that these files
    created.
    """
    generate.generate_files(
        context={'cookiecutter': {'pyhooks': 'pyhooks'}},
        repo_dir='tests/test-pyhooks/',
        output_dir='tests/test-pyhooks/',
    )
    assert os.path.exists('tests/test-pyhooks/inputpyhooks/python_pre.txt')
    assert os.path.exists('tests/test-pyhooks/inputpyhooks/python_post.txt')


@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_run_python_hooks_cwd():
    """Verify pre and post generation python hooks executed and result in current dir.

    Each hook should create in target directory. Test verifies that these files
    created.
    """
    generate.generate_files(
        context={'cookiecutter': {'pyhooks': 'pyhooks'}}, repo_dir='tests/test-pyhooks/'
    )
    assert os.path.exists('inputpyhooks/python_pre.txt')
    assert os.path.exists('inputpyhooks/python_post.txt')


@pytest.mark.skipif(WINDOWS, reason='OSError.errno=8 is not thrown on Windows')
@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_empty_hooks():
    """Verify error is raised on empty hook script. Ignored on windows.

    OSError.errno=8 is not thrown on Windows when the script is empty
    because it always runs through shell instead of needing a shebang.
    """
    with pytest.raises(FailedHookException) as excinfo:
        generate.generate_files(
            context={'cookiecutter': {'shellhooks': 'shellhooks'}},
            repo_dir='tests/test-shellhooks-empty/',
            overwrite_if_exists=True,
        )
    assert 'shebang' in str(excinfo.value)


@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_oserror_hooks(mocker):
    """Verify script error passed correctly to cookiecutter error.

    Here subprocess.Popen function mocked, ie we do not call hook script,
    just produce expected error.
    """
    message = 'Out of memory'

    err = OSError(message)
    err.errno = errno.ENOMEM

    prompt = mocker.patch('subprocess.Popen')
    prompt.side_effect = err

    with pytest.raises(FailedHookException) as excinfo:
        generate.generate_files(
            context={'cookiecutter': {'shellhooks': 'shellhooks'}},
            repo_dir='tests/test-shellhooks-empty/',
            overwrite_if_exists=True,
        )
    assert message in str(excinfo.value)


@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_run_failing_hook_removes_output_directory():
    """Verify project directory not created or removed if hook failed."""
    repo_path = os.path.abspath('tests/test-hooks/')
    hooks_path = os.path.abspath('tests/test-hooks/hooks')

    hook_dir = os.path.join(repo_path, 'hooks')
    template = os.path.join(repo_path, 'input{{cookiecutter.hooks}}')
    os.mkdir(repo_path)
    os.mkdir(hook_dir)
    os.mkdir(template)

    hook_path = os.path.join(hooks_path, 'pre_gen_project.py')

    with open(hook_path, 'w') as f:
        f.write("#!/usr/bin/env python\n")
        f.write("import sys; sys.exit(1)\n")

    with pytest.raises(FailedHookException) as excinfo:
        generate.generate_files(
            context={'cookiecutter': {'hooks': 'hooks'}},
            repo_dir='tests/test-hooks/',
            overwrite_if_exists=True,
        )

    assert 'Hook script failed' in str(excinfo.value)
    assert not os.path.exists('inputhooks')


@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_run_failing_hook_preserves_existing_output_directory():
    """Verify project directory not removed if exist before hook failed."""
    repo_path = os.path.abspath('tests/test-hooks/')
    hooks_path = os.path.abspath('tests/test-hooks/hooks')

    hook_dir = os.path.join(repo_path, 'hooks')
    template = os.path.join(repo_path, 'input{{cookiecutter.hooks}}')
    os.mkdir(repo_path)
    os.mkdir(hook_dir)
    os.mkdir(template)

    hook_path = os.path.join(hooks_path, 'pre_gen_project.py')

    with open(hook_path, 'w') as f:
        f.write("#!/usr/bin/env python\n")
        f.write("import sys; sys.exit(1)\n")

    os.mkdir('inputhooks')
    with pytest.raises(FailedHookException) as excinfo:
        generate.generate_files(
            context={'cookiecutter': {'hooks': 'hooks'}},
            repo_dir='tests/test-hooks/',
            overwrite_if_exists=True,
        )

    assert 'Hook script failed' in str(excinfo.value)
    assert os.path.exists('inputhooks')


@pytest.mark.skipif(sys.platform.startswith('win'), reason="Linux only test")
@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_run_shell_hooks(tmpdir):
    """Verify pre and post generate project shell hooks executed.

    This test for .sh files.
    """
    generate.generate_files(
        context={'cookiecutter': {'shellhooks': 'shellhooks'}},
        repo_dir='tests/test-shellhooks/',
        output_dir=os.path.join(str(tmpdir), 'test-shellhooks'),
    )
    shell_pre_file = os.path.join(
        str(tmpdir), 'test-shellhooks', 'inputshellhooks', 'shell_pre.txt'
    )
    shell_post_file = os.path.join(
        str(tmpdir), 'test-shellhooks', 'inputshellhooks', 'shell_post.txt'
    )
    assert os.path.exists(shell_pre_file)
    assert os.path.exists(shell_post_file)


@pytest.mark.skipif(not sys.platform.startswith('win'), reason="Win only test")
@pytest.mark.usefixtures('clean_system', 'remove_additional_folders')
def test_run_shell_hooks_win(tmpdir):
    """Verify pre and post generate project shell hooks executed.

    This test for .bat files.
    """
    generate.generate_files(
        context={'cookiecutter': {'shellhooks': 'shellhooks'}},
        repo_dir='tests/test-shellhooks-win/',
        output_dir=os.path.join(str(tmpdir), 'test-shellhooks-win'),
    )
    shell_pre_file = os.path.join(
        str(tmpdir), 'test-shellhooks-win', 'inputshellhooks', 'shell_pre.txt'
    )
    shell_post_file = os.path.join(
        str(tmpdir), 'test-shellhooks-win', 'inputshellhooks', 'shell_post.txt'
    )
    assert os.path.exists(shell_pre_file)
    assert os.path.exists(shell_post_file)



# -*- coding: utf-8 -*-

"""Functions for prompting the user for project info."""

from collections import OrderedDict
import json

import click

from jinja2.exceptions import UndefinedError

from cookiecutter.exceptions import UndefinedVariableInTemplate
from cookiecutter.environment import StrictEnvironment, render_variable
from cookiecutter.operator import parse_operator


def read_user_variable(var_name, default_value):
    """Prompt user for variable and return the entered value or given default.

    :param str var_name: Variable of the context to query the user
    :param default_value: Value that will be returned if no input happens
    """
    # Please see https://click.palletsprojects.com/en/7.x/api/#click.prompt
    return click.prompt(var_name, default=default_value)


def read_user_yes_no(question, default_value):
    """Prompt the user to reply with 'yes' or 'no' (or equivalent values).

    Note:
      Possible choices are 'true', '1', 'yes', 'y' or 'false', '0', 'no', 'n'

    :param str question: Question to the user
    :param default_value: Value that will be returned if no input happens
    """
    # Please see https://click.palletsprojects.com/en/7.x/api/#click.prompt
    return click.prompt(question, default=default_value, type=click.BOOL)


def read_repo_password(question):
    """Prompt the user to enter a password.

    :param str question: Question to the user
    """
    # Please see https://click.palletsprojects.com/en/7.x/api/#click.prompt
    return click.prompt(question, hide_input=True)


def read_user_choice(var_name, options):
    """Prompt the user to choose from several options for the given variable.

    The first item will be returned if no input happens.

    :param str var_name: Variable as specified in the context
    :param list options: Sequence of options that are available to select from
    :return: Exactly one item of ``options`` that has been chosen by the user
    """
    # Please see https://click.palletsprojects.com/en/7.x/api/#click.prompt
    if not isinstance(options, list):
        raise TypeError

    if not options:
        raise ValueError

    choice_map = OrderedDict(
        (u'{}'.format(i), value) for i, value in enumerate(options, 1)
    )
    choices = choice_map.keys()
    default = u'1'

    choice_lines = [u'{} - {}'.format(*c) for c in choice_map.items()]
    prompt = u'\n'.join(
        (
            u'Select {}:'.format(var_name),
            u'\n'.join(choice_lines),
            u'Choose from {}'.format(u', '.join(choices)),
        )
    )

    user_choice = click.prompt(
        prompt, type=click.Choice(choices), default=default, show_choices=False
    )
    return choice_map[user_choice]


def process_json(user_value):
    """Load user-supplied value as a JSON dict.

    :param str user_value: User-supplied value to load as a JSON dict
    """
    try:
        user_dict = json.loads(user_value, object_pairs_hook=OrderedDict)
    except Exception:
        # Leave it up to click to ask the user again
        raise click.UsageError('Unable to decode to JSON.')

    if not isinstance(user_dict, dict):
        # Leave it up to click to ask the user again
        raise click.UsageError('Requires JSON dict.')

    return user_dict


def read_user_dict(var_name, default_value):
    """Prompt the user to provide a dictionary of data.

    :param str var_name: Variable as specified in the context
    :param default_value: Value that will be returned if no input is provided
    :return: A Python dictionary to use in the context.
    """
    # Please see https://click.palletsprojects.com/en/7.x/api/#click.prompt
    if not isinstance(default_value, dict):
        raise TypeError

    default_display = 'default'

    user_value = click.prompt(
        var_name, default=default_display, type=click.STRING, value_proc=process_json
    )

    if user_value == default_display:
        # Return the given default w/o any processing
        return default_value
    return user_value


def prompt_choice_for_config(
    cookiecutter_dict, env, key, options, no_input, context_key
):
    """Prompt user with a set of options to choose from.

    Each of the possible choices is rendered beforehand.
    """
    rendered_options = [
        render_variable(env, raw, cookiecutter_dict, context_key) for raw in options
    ]

    if no_input:
        return rendered_options[0]
    return read_user_choice(key, rendered_options)


def prompt_for_config(context, no_input=False, context_key=None):
    """Prompt user to enter a new config.

    :param dict context: Source for field names and sample values.
    :param no_input: Prompt the user at command line for manual configuration?
    """
    cookiecutter_dict = OrderedDict([])
    env = StrictEnvironment(context=context)

    if not context_key:
        context_key = next(iter(context))

    # First pass: Handle simple and raw variables, plus choices.
    # These must be done first because the dictionaries keys and
    # values might refer to them.
    for key, raw in context[context_key].items():
        if key.startswith(u'_'):
            cookiecutter_dict[key] = raw
            continue

        try:
            if isinstance(raw, list):
                # We are dealing with a choice variable
                val = prompt_choice_for_config(
                    cookiecutter_dict, env, key, raw, no_input, context_key
                )
                cookiecutter_dict[key] = val
            elif not isinstance(raw, dict):
                # We are dealing with a regular variable
                val = render_variable(env, raw, cookiecutter_dict, context_key)

                if not no_input:
                    val = read_user_variable(key, val)

                cookiecutter_dict[key] = val
        except UndefinedError as err:
            msg = "Unable to render variable '{}'".format(key)
            raise UndefinedVariableInTemplate(msg, err, context)

    # Second pass; handle the dictionaries.
    for key, raw in context[context_key].items():
        try:
            if isinstance(raw, dict):
                # dict parsing logic
                if 'type' not in raw:
                    val = render_variable(env, raw, cookiecutter_dict, context_key)
                    if not no_input:
                        val = read_user_dict(key, val)
                    cookiecutter_dict[key] = val
                else:
                    cookiecutter_dict = parse_operator(
                        context,
                        key,
                        dict(cookiecutter_dict),
                        no_input=no_input,
                        context_key=context_key,
                    )

        except UndefinedError as err:
            msg = "Unable to render variable '{}'".format(key)
            raise UndefinedVariableInTemplate(msg, err, context)

    return cookiecutter_dict



# -*- coding: utf-8 -*-

"""Tests around prompting for and handling of choice variables."""

import click
import pytest

from cookiecutter.prompt import read_user_choice

OPTIONS = ['hello', 'world', 'foo', 'bar']

EXPECTED_PROMPT = """Select varname:
1 - hello
2 - world
3 - foo
4 - bar
Choose from 1, 2, 3, 4"""


@pytest.mark.parametrize('user_choice, expected_value', enumerate(OPTIONS, 1))
def test_click_invocation(mocker, user_choice, expected_value):
    """Test click function called correctly by cookiecutter.

    Test for choice type invocation.
    """
    choice = mocker.patch('click.Choice')
    choice.return_value = click.Choice(OPTIONS)

    prompt = mocker.patch('click.prompt')
    prompt.return_value = '{}'.format(user_choice)

    assert read_user_choice('varname', OPTIONS) == expected_value

    prompt.assert_called_once_with(
        EXPECTED_PROMPT, type=click.Choice(OPTIONS), default='1', show_choices=False
    )


def test_raise_if_options_is_not_a_non_empty_list():
    """Test function called by cookiecutter raise expected errors.

    Test for choice type invocation.
    """
    with pytest.raises(TypeError):
        read_user_choice('foo', 'NOT A LIST')

    with pytest.raises(ValueError):
        read_user_choice('foo', [])



# -*- coding: utf-8 -*-

"""Collection of tests around log handling."""

import logging

import pytest

from cookiecutter.log import configure_logger


def create_log_records():
    """Test function, create log entries in expected stage of test."""
    cookiecutter_logger = logging.getLogger('cookiecutter')
    foo_logger = logging.getLogger('cookiecutter.foo')
    foobar_logger = logging.getLogger('cookiecutter.foo.bar')

    cookiecutter_logger.info('Welcome to Cookiecutter')
    cookiecutter_logger.debug('Generating project from pytest-plugin')
    foo_logger.info('Loading user config from home dir')
    foobar_logger.debug("I don't know.")
    foobar_logger.debug('I wanted to save the world.')
    foo_logger.error('Aw, snap! Something went wrong')
    cookiecutter_logger.debug('Successfully generated project')


@pytest.fixture
def info_messages():
    """Fixture. List of test info messages."""
    return [
        'INFO: Welcome to Cookiecutter',
        'INFO: Loading user config from home dir',
        'ERROR: Aw, snap! Something went wrong',
    ]


@pytest.fixture
def debug_messages():
    """Fixture. List of test debug messages."""
    return [
        'INFO cookiecutter: ' 'Welcome to Cookiecutter',
        'DEBUG cookiecutter: ' 'Generating project from pytest-plugin',
        'INFO cookiecutter.foo: ' 'Loading user config from home dir',
        "DEBUG cookiecutter.foo.bar: " "I don't know.",
        'DEBUG cookiecutter.foo.bar: ' 'I wanted to save the world.',
        'ERROR cookiecutter.foo: ' 'Aw, snap! Something went wrong',
        'DEBUG cookiecutter: ' 'Successfully generated project',
    ]


@pytest.fixture
def info_logger():
    """Fixture. Call cookiecutter logger setup with `info` debug level."""
    return configure_logger(stream_level='INFO')


@pytest.fixture
def debug_logger():
    """Fixture. Call cookiecutter logger setup with `debug` debug level."""
    return configure_logger(stream_level='DEBUG')


@pytest.fixture
def debug_file(tmpdir):
    """Fixture. Generate debug file location for tests."""
    return tmpdir / 'pytest-plugin.log'


@pytest.fixture
def info_logger_with_file(debug_file):
    """Fixture. Call cookiecutter logger setup with `info` debug level + `file`."""
    return configure_logger(stream_level='INFO', debug_file=str(debug_file))


def test_info_stdout_logging(caplog, info_logger, info_messages):
    """Test that stdout logs use info format and level."""
    [stream_handler] = info_logger.handlers
    assert isinstance(stream_handler, logging.StreamHandler)
    assert stream_handler.level == logging.INFO

    create_log_records()

    stream_messages = [
        stream_handler.format(r)
        for r in caplog.records
        if r.levelno >= stream_handler.level
    ]

    assert stream_messages == info_messages


def test_debug_stdout_logging(caplog, debug_logger, debug_messages):
    """Test that stdout logs use debug format and level."""
    [stream_handler] = debug_logger.handlers
    assert isinstance(stream_handler, logging.StreamHandler)
    assert stream_handler.level == logging.DEBUG

    create_log_records()

    stream_messages = [
        stream_handler.format(r)
        for r in caplog.records
        if r.levelno >= stream_handler.level
    ]

    assert stream_messages == debug_messages


def test_debug_file_logging(caplog, info_logger_with_file, debug_file, debug_messages):
    """Test that logging to stdout uses a different format and level than \
    the the file handler."""
    [file_handler, stream_handler] = info_logger_with_file.handlers
    assert isinstance(file_handler, logging.FileHandler)
    assert isinstance(stream_handler, logging.StreamHandler)
    assert stream_handler.level == logging.INFO
    assert file_handler.level == logging.DEBUG

    create_log_records()

    assert debug_file.exists()

    # Last line in the log file is an empty line
    assert debug_file.readlines(cr=False) == debug_messages + ['']



# -*- coding: utf-8 -*-

"""Test cookiecutter for work without any input.

Tests in this file execute `cookiecutter()` with `no_input=True` flag and
verify result with different settings in `cookiecutter.json`.
"""

import os
import textwrap
import pytest
from _collections import OrderedDict

from cookiecutter import main, utils


@pytest.fixture(scope='function')
def remove_additional_dirs(monkeypatch, request):
    """Fixture. Remove special directories which are created during the tests."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    def fin_remove_additional_dirs():
        if os.path.isdir('fake-project'):
            utils.rmtree('fake-project')
        if os.path.isdir('fake-project-extra'):
            utils.rmtree('fake-project-extra')
        if os.path.isdir('fake-project-templated'):
            utils.rmtree('fake-project-templated')
        if os.path.isdir('fake-project-dict'):
            utils.rmtree('fake-project-dict')
        if os.path.isdir('fake-tmp'):
            utils.rmtree('fake-tmp')

    request.addfinalizer(fin_remove_additional_dirs)


# Parameter messing up cleanup and confusing things
# @pytest.mark.parametrize('path', ['fake-repo-pre/', 'fake-repo-pre'])
@pytest.mark.usefixtures('clean_system', 'remove_additional_dirs')
def test_cookiecutter_no_input_return_project_dir(monkeypatch):
    """Verify `cookiecutter` create project dir on input with or without slash."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    context = main.cookiecutter('fake-repo-pre', no_input=True)

    assert type(context) == OrderedDict
    assert os.path.isdir('fake-repo-pre/{{cookiecutter.repo_name}}')
    assert not os.path.isdir('fake-repo-pre/fake-project')

    # project_dir = os.path.join(os.path.abspath(os.path.curdir), 'fake-project')
    # x = os.listdir(project_dir)
    # y = os.path.isdir(project_dir)
    # Derrr,,,, I don't see what is wrong here but it works as expected?
    # assert os.path.isdir(project_dir)
    # assert os.path.isfile(os.path.join(project_dir, '/README.rst'))
    # assert not os.path.exists('fake-project/json/')


@pytest.mark.usefixtures('clean_system', 'remove_additional_dirs')
def test_cookiecutter_no_input_extra_context(monkeypatch):
    """Verify `cookiecutter` accept `extra_context` argument."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))
    main.cookiecutter(
        'fake-repo-pre',
        no_input=True,
        extra_context={'repo_name': 'fake-project-extra'},
    )
    assert os.path.isdir('fake-project-extra')


@pytest.mark.usefixtures('clean_system', 'remove_additional_dirs')
def test_cookiecutter_templated_context(monkeypatch):
    """Verify Jinja2 templating correctly works in `cookiecutter.json` file."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    main.cookiecutter('fake-repo-tmpl', no_input=True)
    assert os.path.isdir('fake-project-templated')


@pytest.mark.usefixtures('clean_system', 'remove_additional_dirs')
def test_cookiecutter_no_input_return_rendered_file(monkeypatch):
    """Verify Jinja2 templating correctly works in `cookiecutter.json` file."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    context = main.cookiecutter('fake-repo-pre', no_input=True)
    project_dir = os.path.join(os.path.abspath(os.path.curdir), 'fake-project')

    assert type(context) == OrderedDict
    assert project_dir == os.path.abspath('fake-project')
    with open(os.path.join(project_dir, 'README.rst')) as fh:
        contents = fh.read()
    assert "Fake Project" in contents


@pytest.mark.usefixtures('clean_system', 'remove_additional_dirs')
def test_cookiecutter_dict_values_in_context(monkeypatch):
    """Verify configured dictionary from `cookiecutter.json` correctly unpacked."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    context = main.cookiecutter('fake-repo-dict', no_input=True)
    project_dir = os.path.abspath(os.path.curdir)

    assert 'fake-project-dict' in os.listdir(project_dir)
    assert type(context) == OrderedDict

    project_dir = os.path.join(project_dir, 'fake-project-dict')

    with open(os.path.join(project_dir, 'README.md')) as fh:
        contents = fh.read()

    assert (
        contents
        == textwrap.dedent(
            """
        # README


        <dl>
          <dt>Format name:</dt>
          <dd>Bitmap</dd>

          <dt>Extension:</dt>
          <dd>bmp</dd>

          <dt>Applications:</dt>
          <dd>
              <ul>
              <li>Paint</li>
              <li>GIMP</li>
              </ul>
          </dd>
        </dl>

        <dl>
          <dt>Format name:</dt>
          <dd>Portable Network Graphic</dd>

          <dt>Extension:</dt>
          <dd>png</dd>

          <dt>Applications:</dt>
          <dd>
              <ul>
              <li>GIMP</li>
              </ul>
          </dd>
        </dl>

    """
        ).lstrip()
    )


@pytest.mark.usefixtures('clean_system', 'remove_additional_dirs')
def test_cookiecutter_template_cleanup(monkeypatch, mocker):
    """Verify temporary folder for zip unpacking dropped."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    mocker.patch('tempfile.mkdtemp', return_value='fake-tmp', autospec=True)

    mocker.patch(
        'cookiecutter.utils.prompt_and_delete', return_value=True, autospec=True
    )

    main.cookiecutter('files/fake-repo-tmpl.zip', no_input=True)
    assert os.path.isdir('fake-project-templated')

    # The tmp directory will still exist, but the
    # extracted template directory *in* the temp directory will not.
    assert os.path.exists('fake-tmp')
    assert not os.path.exists('fake-tmp/fake-repo-tmpl')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Simple pre-gen hook for testing project folder and custom file creation."""

from __future__ import print_function

print('pre generation hook')
f = open('python_pre.txt', 'w')
f.close()



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.prompt` module."""
import os
from _collections import OrderedDict

from cookiecutter.operator import run_operator, parse_operator

base_name = os.path.abspath(os.path.dirname(__file__))

# TODO: This does not work in testing but does when you run it live
#  Fix this....
context = {
    'cookiecutter': {
        'project_name': "Slartibartfast",
        'details': {"type": "nukikata", "template": base_name, "no_input": True},
    },
}

context_prompt = {
    'cookiecutter': {
        'project_name': "Slartibartfast",
        'details': {"type": "nukikata_prompt", "context": context, "no_input": True},
    },
}


def test_nukikata_operator():
    """Verify simplest functionality."""
    # TODO: Get this to run in IDE - import error
    operator_output, delayed_output = run_operator(
        context['cookiecutter']['details'], context, no_input=True
    )

    assert type(operator_output) == OrderedDict
    assert not delayed_output
    prompt_output = parse_operator(context, 'details', {}, no_input=True)
    assert type(prompt_output) == dict


def test_nukikata_prompt_operator():
    """Verify simplest functionality."""
    operator_output, delayed_output = run_operator(
        context_prompt['cookiecutter']['details'], context_prompt, no_input=True
    )

    assert type(operator_output) == OrderedDict
    assert not delayed_output
    prompt_output = parse_operator(context_prompt, 'details', {}, no_input=True)
    assert type(prompt_output) == dict



# -*- coding: utf-8 -*-

"""Tests for `cookiecutter.utils` module."""

import os
import stat
import sys

import pytest

from cookiecutter import utils


def make_readonly(path):
    """Change the access permissions to readonly for a given file."""
    mode = os.stat(path).st_mode
    os.chmod(path, mode & ~stat.S_IWRITE)


def test_rmtree():
    """Verify `utils.rmtree` remove files marked as read-only."""
    os.mkdir('foo')
    with open('foo/bar', "w") as f:
        f.write("Test data")
    make_readonly('foo/bar')
    utils.rmtree('foo')
    assert not os.path.exists('foo')


def test_make_sure_path_exists():
    """Verify correct True/False response from `utils.make_sure_path_exists`.

    Should return True if directory exist or created.
    Should return False if impossible to create directory (for example protected)
    """
    if sys.platform.startswith('win'):
        existing_directory = os.path.abspath(os.curdir)
        uncreatable_directory = 'a*b'
    else:
        existing_directory = '/usr/'
        uncreatable_directory = '/this-doesnt-exist-and-cant-be-created/'

    assert utils.make_sure_path_exists(existing_directory)
    assert utils.make_sure_path_exists('tests/blah')
    assert utils.make_sure_path_exists('tests/trailingslash/')
    assert not utils.make_sure_path_exists(uncreatable_directory)
    utils.rmtree('tests/blah/')
    utils.rmtree('tests/trailingslash/')


def test_workin():
    """Verify returning to original folder after `utils.work_in` use."""
    cwd = os.getcwd()
    ch_to = 'tests/files'

    class TestException(Exception):
        pass

    def test_work_in():
        with utils.work_in(ch_to):
            test_dir = os.path.join(cwd, ch_to).replace("/", os.sep)
            assert test_dir == os.getcwd()
            raise TestException()

    # Make sure we return to the correct folder
    assert cwd == os.getcwd()

    # Make sure that exceptions are still bubbled up
    with pytest.raises(TestException):
        test_work_in()


def test_prompt_should_ask_and_rm_repo_dir(mocker, tmpdir):
    """In `prompt_and_delete()`, if the user agrees to delete/reclone the \
    repo, the repo should be deleted."""
    mock_read_user = mocker.patch(
        'cookiecutter.utils.read_user_yes_no', return_value=True, autospec=True
    )
    repo_dir = tmpdir.mkdir('repo')

    deleted = utils.prompt_and_delete(str(repo_dir))

    assert mock_read_user.called
    assert not repo_dir.exists()
    assert deleted


def test_prompt_should_ask_and_exit_on_user_no_answer(mocker, tmpdir):
    """In `prompt_and_delete()`, if the user decline to delete/reclone the \
    repo, cookiecutter should exit."""
    mock_read_user = mocker.patch(
        'cookiecutter.utils.read_user_yes_no', return_value=False,
    )
    mock_sys_exit = mocker.patch('sys.exit', return_value=True)
    repo_dir = tmpdir.mkdir('repo')

    deleted = utils.prompt_and_delete(str(repo_dir))

    assert mock_read_user.called
    assert repo_dir.exists()
    assert not deleted
    assert mock_sys_exit.called


def test_prompt_should_ask_and_rm_repo_file(mocker, tmpdir):
    """In `prompt_and_delete()`, if the user agrees to delete/reclone a \
    repo file, the repo should be deleted."""
    mock_read_user = mocker.patch(
        'cookiecutter.utils.read_user_yes_no', return_value=True, autospec=True
    )

    repo_file = tmpdir.join('repo.zip')
    repo_file.write('this is zipfile content')

    deleted = utils.prompt_and_delete(str(repo_file))

    assert mock_read_user.called
    assert not repo_file.exists()
    assert deleted


def test_prompt_should_ask_and_keep_repo_on_no_reuse(mocker, tmpdir):
    """In `prompt_and_delete()`, if the user wants to keep their old \
    cloned template repo, it should not be deleted."""
    mock_read_user = mocker.patch(
        'cookiecutter.utils.read_user_yes_no', return_value=False, autospec=True
    )
    repo_dir = tmpdir.mkdir('repo')

    with pytest.raises(SystemExit):
        utils.prompt_and_delete(str(repo_dir))

    assert mock_read_user.called
    assert repo_dir.exists()


def test_prompt_should_ask_and_keep_repo_on_reuse(mocker, tmpdir):
    """In `prompt_and_delete()`, if the user wants to keep their old \
    cloned template repo, it should not be deleted."""

    def answer(question, default):
        if 'okay to delete' in question:
            return False
        else:
            return True

    mock_read_user = mocker.patch(
        'cookiecutter.utils.read_user_yes_no', side_effect=answer, autospec=True
    )
    repo_dir = tmpdir.mkdir('repo')

    deleted = utils.prompt_and_delete(str(repo_dir))

    assert mock_read_user.called
    assert repo_dir.exists()
    assert not deleted


def test_prompt_should_not_ask_if_no_input_and_rm_repo_dir(mocker, tmpdir):
    """Prompt should not ask if no input and rm dir.

    In `prompt_and_delete()`, if `no_input` is True, the call to
    `prompt.read_user_yes_no()` should be suppressed.
    """
    mock_read_user = mocker.patch(
        'cookiecutter.prompt.read_user_yes_no', return_value=True, autospec=True
    )
    repo_dir = tmpdir.mkdir('repo')

    deleted = utils.prompt_and_delete(str(repo_dir), no_input=True)

    assert not mock_read_user.called
    assert not repo_dir.exists()
    assert deleted


def test_prompt_should_not_ask_if_no_input_and_rm_repo_file(mocker, tmpdir):
    """Prompt should not ask if no input and rm file.

    In `prompt_and_delete()`, if `no_input` is True, the call to
    `prompt.read_user_yes_no()` should be suppressed.
    """
    mock_read_user = mocker.patch(
        'cookiecutter.prompt.read_user_yes_no', return_value=True, autospec=True
    )

    repo_file = tmpdir.join('repo.zip')
    repo_file.write('this is zipfile content')

    deleted = utils.prompt_and_delete(str(repo_file), no_input=True)

    assert not mock_read_user.called
    assert not repo_file.exists()
    assert deleted


@pytest.mark.parametrize(
    'valid_config_file',
    (
        [
            'tests/fixtures/valid/nuki-input/nuki.yml',
            'tests/fixtures/valid/yaml-input/cookiecutter.yaml',
        ]
    ),
)
def test_valid_read_config_file(valid_config_file):
    """Validate generic reader works properly."""
    output = utils.read_config_file(valid_config_file)
    assert output == {'project_slug': 'best_eva', 'stuff': 'things'}



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.prompt` module."""
import pytest
from cookiecutter.operator import run_operator, parse_operator

context = {
    'cookiecutter': {
        'project_name': "Slartibartfast",
        'string': {"type": "split", "items": "stuff.thing"},
        'string-sep': {"type": "split", "items": "stuff-thing", "separator": "-"},
        'list-sep': {
            "type": "split",
            "items": ["stuff-thing", "foo-bar"],
            "separator": "-",
        },
    }
}


@pytest.mark.parametrize(
    'key, expected_value',
    [
        ('string', ['stuff', 'thing']),
        ('string-sep', ['stuff', 'thing']),
        ('list-sep', [['stuff', 'thing'], ['foo', 'bar']]),
    ],
)
def test_split_operator_string(key, expected_value):
    """Verify simplest functionality."""
    operator_output, delayed_output = run_operator(
        context['cookiecutter'][key], context
    )

    assert operator_output == expected_value
    cookiecutter_dict = parse_operator(context, key, {})
    assert cookiecutter_dict == {key: expected_value}



# -*- coding: utf-8 -*-

"""Operator plugin that inherits a base class and is made available through `type`."""
from __future__ import unicode_literals
from __future__ import print_function

import logging
from PyInquirer import prompt

from cookiecutter.operators import BaseOperator

logger = logging.getLogger(__name__)


class InquirerExpandOperator(BaseOperator):
    """Operator for PyInquirer `expand` type prompt."""

    type = 'expand'

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize PyInquirer operator."""  # noqa
        super(InquirerExpandOperator, self).__init__(
            operator_dict=operator_dict, context=context, no_input=no_input
        )

    def execute(self):
        """Run the prompt."""  # noqa
        if not self.no_input:
            if 'name' not in self.operator_dict:
                self.operator_dict.update({'name': 'tmp'})
                return prompt([self.operator_dict])['tmp']
            else:
                return prompt([self.operator_dict])
        elif 'default' in self.operator_dict:
            return self.operator_dict['default']
        else:
            # When no_input then return empty list
            return []



# -*- coding: utf-8 -*-

"""Operator plugin that inherits a base class and is made available through `type`."""
from __future__ import unicode_literals
from __future__ import print_function

import logging

from cookiecutter.operators import BaseOperator

logger = logging.getLogger(__name__)


class SplitOperator(BaseOperator):
    """Operator for PyInquirer type prompts."""

    type = 'split'

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize operator."""
        super(SplitOperator, self).__init__(
            operator_dict=operator_dict, context=context, no_input=no_input
        )
        # Defaulting to run inline
        self.post_gen_operator = (
            self.operator_dict['delay'] if 'delay' in self.operator_dict else False
        )
        self.separator = (
            self.operator_dict['separator']
            if 'separator' in self.operator_dict
            else "."
        )

    def execute(self):
        """Run the operator."""  # noqa
        if isinstance(self.operator_dict['items'], str):
            # If item is a string then return a list
            return self.operator_dict['items'].split(self.separator)
        elif isinstance(self.operator_dict['items'], list):
            # If items is a list then return a nested list
            output = []
            for i in self.operator_dict['items']:
                output.append(i.split('-'))
            return output
        else:
            raise NotImplementedError(
                "Have not implemented dict input to " "`items` for `type` 'split'"
            )



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.prompt` module."""
import os

from cookiecutter.operator import run_operator, parse_operator

base_dir = os.path.dirname(__file__)

context = {
    'cookiecutter': {
        'project_name': "Slartibartfast",
        'details': {"type": "command", "command": "ls " + base_dir},
    }
}


def test_command():
    """Verify simplest functionality."""
    if os.name == 'nt':
        # Not testing windows
        pass

    else:
        operator_output, delayed_output = run_operator(
            context['cookiecutter']['details'], context
        )
        expected_output = """__init__.py
__pycache__
test_command.py
"""
        assert operator_output == expected_output
        assert not delayed_output

        cookiecutter_dict = parse_operator(context, 'details', {})
        assert cookiecutter_dict == {'details': expected_output}



# -*- coding: utf-8 -*-

"""Tests around using locally cached cookiecutter template repositories."""

import os

import pytest

from cookiecutter import repository, exceptions


def test_finds_local_repo(monkeypatch, tmpdir):
    """A valid local repository should be returned."""
    test_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..')
    monkeypatch.chdir(test_dir)

    project_dir, context_file, cleanup = repository.determine_repo_dir(
        'tests/fake-repo',
        abbreviations={},
        clone_to_dir=str(tmpdir),
        checkout=None,
        no_input=True,
    )

    assert 'tests/fake-repo' == project_dir
    assert context_file == 'cookiecutter.json'
    assert not cleanup


def test_local_repo_with_no_context_raises(tmpdir):
    """A local repository without a cookiecutter.json should raise a \
    `RepositoryNotFound` exception."""
    template_path = os.path.join('tests', 'fake-repo-bad')
    with pytest.raises(exceptions.RepositoryNotFound) as err:
        repository.determine_repo_dir(
            template_path,
            abbreviations={},
            clone_to_dir=str(tmpdir),
            checkout=None,
            no_input=True,
        )

    assert str(err.value) == (
        'A valid repository for "{}" could not be found in the following '
        'locations:\n{}'.format(
            template_path,
            '\n'.join([template_path, str(tmpdir / 'tests/fake-repo-bad')]),
        )
    )


def test_local_repo_typo(tmpdir):
    """An unknown local repository should raise a `RepositoryNotFound` \
    exception."""
    template_path = os.path.join('tests', 'unknown-repo')
    with pytest.raises(exceptions.RepositoryNotFound) as err:
        repository.determine_repo_dir(
            template_path,
            abbreviations={},
            clone_to_dir=str(tmpdir),
            checkout=None,
            no_input=True,
        )

    assert str(err.value) == (
        'A valid repository for "{}" could not be found in the following '
        'locations:\n{}'.format(
            template_path,
            '\n'.join([template_path, str(tmpdir / 'tests/unknown-repo')]),
        )
    )



# -*- coding: utf-8 -*-

"""Module for setting up logging."""

import logging
import sys

LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL,
}

LOG_FORMATS = {
    'DEBUG': u'%(levelname)s %(name)s: %(message)s',
    'INFO': u'%(levelname)s: %(message)s',
}


def configure_logger(stream_level='DEBUG', debug_file=None):
    """Configure logging for cookiecutter.

    Set up logging to stdout with given level. If ``debug_file`` is given set
    up logging to file with DEBUG level.
    """
    # Set up 'cookiecutter' logger
    logger = logging.getLogger('cookiecutter')
    logger.setLevel(logging.DEBUG)

    # Remove all attached handlers, in case there was
    # a logger with using the name 'cookiecutter'
    del logger.handlers[:]

    # Create a file handler if a log file is provided
    if debug_file is not None:
        debug_formatter = logging.Formatter(LOG_FORMATS['DEBUG'])
        file_handler = logging.FileHandler(debug_file)
        file_handler.setLevel(LOG_LEVELS['DEBUG'])
        file_handler.setFormatter(debug_formatter)
        logger.addHandler(file_handler)

    # Get settings based on the given stream_level
    log_formatter = logging.Formatter(LOG_FORMATS[stream_level])
    log_level = LOG_LEVELS[stream_level]

    # Create a stream handler
    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setLevel(log_level)
    stream_handler.setFormatter(log_formatter)
    logger.addHandler(stream_handler)

    return logger



# -*- coding: utf-8 -*-

"""Jinja2 environment and extensions loading."""
import ast
import re

import six
from jinja2 import Environment, StrictUndefined

from cookiecutter.exceptions import UnknownExtension


class ExtensionLoaderMixin(object):
    """Mixin providing sane loading of extensions specified in a given context.

    The context is being extracted from the keyword arguments before calling
    the next parent class in line of the child.
    """

    def __init__(self, **kwargs):
        """Initialize the Jinja2 Environment object while loading extensions.

        Does the following:

        1. Establishes default_extensions (currently just a Time feature)
        2. Reads extensions set in the cookiecutter.json _extensions key.
        3. Attempts to load the extensions. Provides useful error if fails.
        """
        context = kwargs.pop('context', {})

        default_extensions = [
            'cookiecutter.extensions.JsonifyExtension',
            'cookiecutter.extensions.RandomStringExtension',
            'cookiecutter.extensions.SlugifyExtension',
            'jinja2_time.TimeExtension',
        ]
        extensions = default_extensions + self._read_extensions(context)

        try:
            super(ExtensionLoaderMixin, self).__init__(extensions=extensions, **kwargs)
        except ImportError as err:
            raise UnknownExtension('Unable to load extension: {}'.format(err))

    def _read_extensions(self, context):
        """Return list of extensions as str to be passed on to the Jinja2 env.

        If context does not contain the relevant info, return an empty
        list instead.
        """
        try:
            extensions = context['cookiecutter']['_extensions']
        except KeyError:
            return []
        else:
            return [str(ext) for ext in extensions]


class StrictEnvironment(ExtensionLoaderMixin, Environment):
    """Create strict Jinja2 environment.

    Jinja2 environment will raise error on undefined variable in template-
    rendering context.
    """

    def __init__(self, **kwargs):
        """Set the standard Cookiecutter StrictEnvironment.

        Also loading extensions defined in cookiecutter.json's _extensions key.
        """
        super(StrictEnvironment, self).__init__(undefined=StrictUndefined, **kwargs)


def render_variable(env, raw, cookiecutter_dict, context_key):
    """Render the next variable to be displayed in the user prompt.

    Inside the prompting taken from the cookiecutter.json file, this renders
    the next variable. For example, if a project_name is "Peanut Butter
    Cookie", the repo_name could be be rendered with:

        `{{ cookiecutter.project_name.replace(" ", "_") }}`.

    This is then presented to the user as the default.

    :param Environment env: A Jinja2 Environment object.
    :param raw: The next value to be prompted for by the user.
    :param dict cookiecutter_dict: The current context as it's gradually
        being populated with variables.
    :return: The rendered value for the default variable.
    """
    if raw is None:
        return None
    elif isinstance(raw, dict):
        return {
            render_variable(env, k, cookiecutter_dict, context_key): render_variable(
                env, v, cookiecutter_dict, context_key
            )
            for k, v in raw.items()
        }
    elif isinstance(raw, list):
        return [render_variable(env, v, cookiecutter_dict, context_key) for v in raw]
    elif not isinstance(raw, six.string_types):
        raw = str(raw)

    template = env.from_string(raw)

    rendered_template = template.render({context_key: cookiecutter_dict})

    LIST_REGEX = r'^\[.*\]$'
    if bool(re.search(LIST_REGEX, rendered_template)):
        """If variable looks like list, return literal list"""
        return ast.literal_eval(rendered_template)

    DICT_REGEX = r'^\{.*\}$'
    if bool(re.search(DICT_REGEX, rendered_template)):
        """If variable looks like dict, return literal dict"""
        return ast.literal_eval(rendered_template)

    return rendered_template



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.prompt` module."""
from cookiecutter.operator import run_operator, parse_operator
import os

base_dir = os.path.dirname(__file__)

context = {
    'cookiecutter': {
        'project_name': "Slartibartfast",
        'details': {"type": "listdir", "directory": os.path.join(base_dir, "dir")},
        'hidden': {
            "type": "listdir",
            "directory": os.path.join(base_dir, "dir"),
            "ignore_hidden_files": True,
        },
        'list_directories': {
            "type": "listdir",
            "directories": [
                os.path.join(base_dir, "dir"),
                os.path.join(base_dir, "dirs"),
            ],
            "ignore_hidden_files": True,
        },
    }
}


def test_listdir_operator(monkeypatch):
    """Verify simplest functionality."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))
    operator_output, delayed_output = run_operator(
        context['cookiecutter']['details'], context
    )

    assert len(operator_output) == 3

    cookiecutter_dict = parse_operator(context, 'details', {})
    assert len(cookiecutter_dict['details']) == 3


def test_listdir_operator_ignore_hidden(monkeypatch):
    """Verify simplest functionality."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    operator_output, delayed_output = run_operator(
        context['cookiecutter']['hidden'], context
    )

    assert len(operator_output) == 2

    cookiecutter_dict = parse_operator(context, 'hidden', {})
    assert len(cookiecutter_dict['hidden']) == 2


# def test_listdir_operator_directories_list(monkeypatch):
#     """Verify simplest functionality."""
#     monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))
# operator_output, delayed_output = run_operator(
#     context['cookiecutter']['list_directories'], context
# )
#
# expected_output = {
#     os.path.join(base_dir, "dir"): ['things.py', 'stuff.txt', '.hidden-stuff'],
#     os.path.join(base_dir, "dirs"): [
#         'dir2',
#         'dir1',
#         'things.py',
#         'stuff.txt',
#         '.hidden-stuff',
#     ],
# }
# Breaking in travis
# assert operator_output == expected_output

# cookiecutter_dict = parse_operator(context, 'list_directories', {})
# assert cookiecutter_dict == {'list_directories': expected_output}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# flake8: noqa

"""Simple post-gen hook for testing the handling of different exit codes."""

import sys

{% if cookiecutter.abort_post_gen == "yes" %}
sys.exit(5)
{% else %}
sys.exit(0)
{% endif %}



# -*- coding: utf-8 -*-

"""Tests for `cookiecutter.prompt` module."""

import platform
from collections import OrderedDict

import pytest
import six

import cookiecutter.environment
from cookiecutter import prompt, exceptions, environment


@pytest.fixture(autouse=True)
def patch_readline_on_win(monkeypatch):
    """Fixture. Overwrite windows end of line to linux standard."""
    if 'windows' in platform.platform().lower():
        monkeypatch.setattr('sys.stdin.readline', lambda: '\n')


class TestRenderVariable:
    """Class to unite simple and complex tests for render_variable function."""

    @pytest.mark.parametrize(
        'raw_var, rendered_var',
        [
            (1, '1'),
            (True, 'True'),
            ('foo', 'foo'),
            ('{{cookiecutter.project}}', 'foobar'),
            (None, None),
        ],
    )
    def test_convert_to_str(self, mocker, raw_var, rendered_var):
        """Verify simple items correctly rendered to strings."""
        env = environment.StrictEnvironment()
        from_string = mocker.patch(
            'cookiecutter.prompt.StrictEnvironment.from_string', wraps=env.from_string
        )
        context = {'project': 'foobar'}

        result = cookiecutter.environment.render_variable(
            env, raw_var, context, context_key='cookiecutter'
        )
        assert result == rendered_var

        # Make sure that non None non str variables are converted beforehand
        if raw_var is not None:
            if not isinstance(raw_var, six.string_types):
                raw_var = str(raw_var)
            from_string.assert_called_once_with(raw_var)
        else:
            assert not from_string.called

    @pytest.mark.parametrize(
        'raw_var, rendered_var',
        [
            ({1: True, 'foo': False}, {'1': 'True', 'foo': 'False'}),
            (
                {'{{cookiecutter.project}}': ['foo', 1], 'bar': False},
                {'foobar': ['foo', '1'], 'bar': 'False'},
            ),
            (['foo', '{{cookiecutter.project}}', None], ['foo', 'foobar', None]),
        ],
    )
    def test_convert_to_str_complex_variables(self, raw_var, rendered_var):
        """Verify tree items correctly rendered."""
        env = environment.StrictEnvironment()
        context = {'project': 'foobar'}

        result = cookiecutter.environment.render_variable(
            env, raw_var, context, context_key='cookiecutter'
        )
        assert result == rendered_var


class TestPrompt(object):
    """Class to unite user prompt related tests."""

    @pytest.mark.parametrize(
        'context',
        [
            {'cookiecutter': {'full_name': 'Your Name'}},
            {'cookiecutter': {'full_name': u'Řekni či napiš své jméno'}},
        ],
        ids=['ASCII default prompt/input', 'Unicode default prompt/input'],
    )
    def test_prompt_for_config(self, monkeypatch, context):
        """Verify `prompt_for_config` call `read_user_variable` on text request."""
        monkeypatch.setattr(
            'cookiecutter.prompt.read_user_variable', lambda var, default: default,
        )

        cookiecutter_dict = prompt.prompt_for_config(context)
        assert cookiecutter_dict == context['cookiecutter']

    def test_prompt_for_config_dict(self, monkeypatch):
        """Verify `prompt_for_config` call `read_user_variable` on dict request."""
        monkeypatch.setattr(
            'cookiecutter.prompt.read_user_dict',
            lambda var, default: {"key": "value", "integer": 37},
        )
        context = {'cookiecutter': {'details': {}}}

        cookiecutter_dict = prompt.prompt_for_config(context)
        assert cookiecutter_dict == {'details': {'key': u'value', 'integer': 37}}

    def test_should_render_dict(self):
        """Verify template inside dictionary variable rendered."""
        context = {
            'cookiecutter': {
                'project_name': 'Slartibartfast',
                'details': {
                    '{{cookiecutter.project_name}}': '{{cookiecutter.project_name}}'
                },
            }
        }

        cookiecutter_dict = prompt.prompt_for_config(context, no_input=True)
        assert cookiecutter_dict == {
            'project_name': 'Slartibartfast',
            'details': {'Slartibartfast': u'Slartibartfast'},
        }

    def test_should_render_deep_dict(self):
        """Verify nested structures like dict in dict, rendered correctly."""
        context = {
            'cookiecutter': {
                'project_name': "Slartibartfast",
                'details': {
                    "key": "value",
                    "integer_key": 37,
                    "other_name": '{{cookiecutter.project_name}}',
                    "dict_key": {
                        "deep_key": "deep_value",
                        "deep_integer": 42,
                        "deep_other_name": '{{cookiecutter.project_name}}',
                        "deep_list": [
                            "deep value 1",
                            "{{cookiecutter.project_name}}",
                            "deep value 3",
                        ],
                    },
                    "list_key": [
                        "value 1",
                        "{{cookiecutter.project_name}}",
                        "value 3",
                    ],
                },
            }
        }

        cookiecutter_dict = prompt.prompt_for_config(context, no_input=True)
        assert cookiecutter_dict == {
            'project_name': "Slartibartfast",
            'details': {
                "key": "value",
                "integer_key": "37",
                "other_name": "Slartibartfast",
                "dict_key": {
                    "deep_key": "deep_value",
                    "deep_integer": "42",
                    "deep_other_name": "Slartibartfast",
                    "deep_list": ["deep value 1", "Slartibartfast", "deep value 3"],
                },
                "list_key": ["value 1", "Slartibartfast", "value 3"],
            },
        }

    def test_prompt_for_templated_config(self, monkeypatch):
        """Verify Jinja2 templating works in unicode prompts."""
        monkeypatch.setattr(
            'cookiecutter.prompt.read_user_variable', lambda var, default: default
        )
        context = {
            'cookiecutter': OrderedDict(
                [
                    ('project_name', u'A New Project'),
                    (
                        'pkg_name',
                        u'{{ cookiecutter.project_name|lower|replace(" ", "") }}',
                    ),
                ]
            )
        }

        exp_cookiecutter_dict = {
            'project_name': u'A New Project',
            'pkg_name': u'anewproject',
        }
        cookiecutter_dict = prompt.prompt_for_config(context)
        assert cookiecutter_dict == exp_cookiecutter_dict

    def test_dont_prompt_for_private_context_var(self, monkeypatch):
        """Verify `read_user_variable` not called for private context variables."""
        monkeypatch.setattr(
            'cookiecutter.prompt.read_user_variable',
            lambda var, default: pytest.fail(
                'Should not try to read a response for private context var'
            ),
        )
        context = {'cookiecutter': {'_copy_without_render': ['*.html']}}
        cookiecutter_dict = prompt.prompt_for_config(context)
        assert cookiecutter_dict == {'_copy_without_render': ['*.html']}

    def test_should_not_render_private_variables(self):
        """Verify private(underscored) variables not rendered by `prompt_for_config`.

        Private variables designed to be raw, same as context input.
        """
        context = {
            'cookiecutter': {
                'project_name': 'Skip render',
                '_skip_jinja_template': '{{cookiecutter.project_name}}',
                '_skip_float': 123.25,
                '_skip_integer': 123,
                '_skip_boolean': True,
                '_skip_nested': True,
            }
        }
        cookiecutter_dict = prompt.prompt_for_config(context, no_input=True)
        assert cookiecutter_dict == context['cookiecutter']


class TestReadUserChoice(object):
    """Class to unite choices prompt related tests."""

    def test_should_invoke_read_user_choice(self, mocker):
        """Verify correct function called for select(list) variables."""
        prompt_choice = mocker.patch(
            'cookiecutter.prompt.prompt_choice_for_config',
            wraps=prompt.prompt_choice_for_config,
        )

        read_user_choice = mocker.patch('cookiecutter.prompt.read_user_choice')
        read_user_choice.return_value = 'all'

        read_user_variable = mocker.patch('cookiecutter.prompt.read_user_variable')

        choices = ['landscape', 'portrait', 'all']
        context = {'cookiecutter': {'orientation': choices}}

        cookiecutter_dict = prompt.prompt_for_config(context)

        assert not read_user_variable.called
        assert prompt_choice.called
        read_user_choice.assert_called_once_with('orientation', choices)
        assert cookiecutter_dict == {'orientation': 'all'}

    def test_should_invoke_read_user_variable(self, mocker):
        """Verify correct function called for string input variables."""
        read_user_variable = mocker.patch('cookiecutter.prompt.read_user_variable')
        read_user_variable.return_value = u'Audrey Roy'

        prompt_choice = mocker.patch('cookiecutter.prompt.prompt_choice_for_config')

        read_user_choice = mocker.patch('cookiecutter.prompt.read_user_choice')

        context = {'cookiecutter': {'full_name': 'Your Name'}}

        cookiecutter_dict = prompt.prompt_for_config(context)

        assert not prompt_choice.called
        assert not read_user_choice.called
        read_user_variable.assert_called_once_with('full_name', 'Your Name')
        assert cookiecutter_dict == {'full_name': u'Audrey Roy'}

    def test_should_render_choices(self, mocker):
        """Verify Jinja2 templating engine works inside choices variables."""
        read_user_choice = mocker.patch('cookiecutter.prompt.read_user_choice')
        read_user_choice.return_value = u'anewproject'

        read_user_variable = mocker.patch('cookiecutter.prompt.read_user_variable')
        read_user_variable.return_value = u'A New Project'

        rendered_choices = [u'foo', u'anewproject', u'bar']

        context = {
            'cookiecutter': OrderedDict(
                [
                    ('project_name', u'A New Project'),
                    (
                        'pkg_name',
                        [
                            u'foo',
                            u'{{ cookiecutter.project_name|lower|replace(" ", "") }}',
                            u'bar',
                        ],
                    ),
                ]
            )
        }

        expected = {
            'project_name': u'A New Project',
            'pkg_name': u'anewproject',
        }
        cookiecutter_dict = prompt.prompt_for_config(context)

        read_user_variable.assert_called_once_with('project_name', u'A New Project')
        read_user_choice.assert_called_once_with('pkg_name', rendered_choices)
        assert cookiecutter_dict == expected


class TestPromptChoiceForConfig(object):
    """Class to unite choices prompt related tests with config test."""

    @pytest.fixture
    def choices(self):
        """Fixture. Just populate choices variable."""
        return ['landscape', 'portrait', 'all']

    @pytest.fixture
    def context(self, choices):
        """Fixture. Just populate context variable."""
        return {'cookiecutter': {'orientation': choices}}

    def test_should_return_first_option_if_no_input(self, mocker, choices, context):
        """Verify prompt_choice_for_config return first list option on no_input=True."""
        read_user_choice = mocker.patch('cookiecutter.prompt.read_user_choice')

        expected_choice = choices[0]

        actual_choice = prompt.prompt_choice_for_config(
            cookiecutter_dict=context,
            env=environment.StrictEnvironment(),
            key='orientation',
            options=choices,
            no_input=True,  # Suppress user input
            context_key='cookiecutter',
        )

        assert not read_user_choice.called
        assert expected_choice == actual_choice

    def test_should_read_user_choice(self, mocker, choices, context):
        """Verify prompt_choice_for_config return user selection on no_input=False."""
        read_user_choice = mocker.patch('cookiecutter.prompt.read_user_choice')
        read_user_choice.return_value = 'all'

        expected_choice = 'all'

        actual_choice = prompt.prompt_choice_for_config(
            cookiecutter_dict=context,
            env=environment.StrictEnvironment(),
            key='orientation',
            options=choices,
            no_input=False,  # Ask the user for input
            context_key='cookiecutter',
        )
        read_user_choice.assert_called_once_with('orientation', choices)
        assert expected_choice == actual_choice


@pytest.mark.parametrize(
    'context',
    (
        {'cookiecutter': {'foo': '{{cookiecutter.nope}}'}},
        {'cookiecutter': {'foo': ['123', '{{cookiecutter.nope}}', '456']}},
        {'cookiecutter': {'foo': {'{{cookiecutter.nope}}': 'value'}}},
        {'cookiecutter': {'foo': {'key': '{{cookiecutter.nope}}'}}},
    ),
    ids=[
        'Undefined variable in cookiecutter dict',
        'Undefined variable in cookiecutter dict with choices',
        'Undefined variable in cookiecutter dict with dict_key',
        'Undefined variable in cookiecutter dict with key_value',
    ],
)
def test_undefined_variable(context):
    """Verify `prompt.prompt_for_config` raises correct error."""
    with pytest.raises(exceptions.UndefinedVariableInTemplate) as err:
        prompt.prompt_for_config(context, no_input=True)

    error = err.value
    assert error.message == "Unable to render variable 'foo'"
    assert error.context == context



# -*- coding: utf-8 -*-

"""test_read_user_yes_no."""

from __future__ import unicode_literals

import click

from cookiecutter.prompt import read_user_yes_no

QUESTION = 'Is it okay to delete and re-clone it?'
DEFAULT = 'y'


def test_click_invocation(mocker):
    """Test click function called correctly by cookiecutter.

    Test for boolean type invocation.
    """
    prompt = mocker.patch('click.prompt')
    prompt.return_value = DEFAULT

    assert read_user_yes_no(QUESTION, DEFAULT) == DEFAULT

    click.prompt.assert_called_once_with(QUESTION, default=DEFAULT, type=click.BOOL)



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.prompt` module."""

# import pytest
import cookiecutter.operator
from cookiecutter import prompt


# Only tests for the basic logic of the parsing of the operator calls


def test_operator_when_false_no_input_no_default():
    """Verify `prompt_for_config` call `read_user_variable` on dict request."""
    context = {
        'cookiecutter': {
            'project_name': "Slartibartfast",
            'details': {
                "message": "what foo?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"},],
                'when': "{{ cookiecutter.project_name != 'Slartibartfast' }}",
            },
        }
    }

    cookiecutter_dict = {
        'project_name': "Slartibartfast",
    }

    out_dict = cookiecutter.operator.parse_operator(
        context, 'details', cookiecutter_dict, no_input=True
    )
    assert out_dict == cookiecutter_dict


def test_operator_when_true_no_input_no_default():
    """Verify `prompt_for_config` call `read_user_variable` on dict request."""
    context = {
        'cookiecutter': {
            'project_name': "Slartibartfast",
            'details': {
                "message": "what foo?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"},],
                'when': "{{ cookiecutter.project_name == 'Slartibartfast' }}",
            },
        }
    }

    cookiecutter_dict = {
        'project_name': "Slartibartfast",
    }

    out_dict = cookiecutter.operator.parse_operator(
        context, 'details', cookiecutter_dict, no_input=True
    )
    assert out_dict == cookiecutter_dict
    assert out_dict != {'project_name': "Slartibartfast"}


def test_operator_when_true_no_input_with_default():
    """Verify `prompt_for_config` call `read_user_variable` on dict request."""
    context = {
        'cookiecutter': {
            'project_name': "Slartibartfast",
            'details': {
                "message": "what fooss?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"},],
                'when': "{{ cookiecutter.project_name == 'Slartibartfast' }}",
                'default': 'stuff',
            },
        }
    }

    cookiecutter_dict = {
        'project_name': "Slartibartfast",
    }

    out_dict = cookiecutter.operator.parse_operator(
        context, 'details', cookiecutter_dict, no_input=True
    )
    assert out_dict == {'project_name': 'Slartibartfast', 'details': 'stuff'}


def test_operator_when_true_no_input_loop():
    """Verify `prompt_for_config` call `read_user_variable` on dict request."""
    context = {
        'cookiecutter': {
            'details': {
                "message": "what details?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"},],
                'when': "{{ cookiecutter.project_name == 'Slartibartfast' }}",
                'loop': ['foo', 'bar'],
                'default': '{{ cookiecutter.item }}',
            },
        }
    }

    cookiecutter_dict = {
        'project_name': "Slartibartfast",
    }

    out_dict = cookiecutter.operator.parse_operator(
        context, 'details', cookiecutter_dict, no_input=True
    )
    assert out_dict == {'project_name': 'Slartibartfast', 'details': ['foo', 'bar']}


def test_prompt_operator_when_true_no_input_loop():
    """Verify `prompt_for_config` call `read_user_variable` on dict request."""
    context = {
        'cookiecutter': {
            'project_name': "Slartibartfast",
            'details': {
                "message": "what details?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"}],
                'when': "{{ cookiecutter.project_name == 'Slartibartfast' }}",
                'loop': ['foo', 'bar'],
                'default': 'stuff',
            },
            'things': {
                "message": "what things?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"}],
                'when': "{{ 'stuff' in cookiecutter.details }}",
                'default': 'mo tings',
            },
        }
    }

    out_dict = prompt.prompt_for_config(context, no_input=True)
    assert out_dict == {
        'project_name': 'Slartibartfast',
        'details': ['stuff', 'stuff'],
        'things': 'mo tings',
    }


def test_prompt_operator_when_true_no_input_loop_item():
    """Verify `prompt_for_config` call `read_user_variable` on dict request."""
    context = {
        'cookiecutter': {
            'project_name': "Slartibartfast",
            'details': {
                "message": "what details?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"},],
                'when': "{{ cookiecutter.project_name == 'Slartibartfast' }}",
                'loop': ['foo', 'bar'],
                'default': '{{ cookiecutter.item }}',
            },
            'things': {
                "message": "what things?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"},],
                'when': "{{ 'foo' in cookiecutter.details }}",
                'default': 'mo tings',
            },
        }
    }

    out_dict = prompt.prompt_for_config(context, no_input=True)
    assert out_dict == {
        'project_name': 'Slartibartfast',
        'details': ['foo', 'bar'],
        'things': 'mo tings',
    }


def test_prompt_operator_when_true_no_input_loop_list():
    """Verify `prompt_for_config` call `read_user_variable` on dict request."""
    context = {
        'cookiecutter': {
            'project_name': "Slartibartfast",
            'details': {
                "message": "what details?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"},],
                'when': "{{ cookiecutter.project_name == 'Slartibartfast' }}",
                'loop': ['foo', 'bar'],
                'default': '{{ cookiecutter.item }}',
            },
            'things': {
                "message": "what things?",
                "type": "checkbox",
                "choices": [{"name": "value 1"}, {"name": "value 2"},],
                'when': "{{ 'foo' in cookiecutter.details }}",
                'loop': "{{ cookiecutter.details }}",
                'default': '{{ cookiecutter.item }}',
            },
        }
    }

    out_dict = prompt.prompt_for_config(context, no_input=True)
    assert out_dict == {
        'project_name': 'Slartibartfast',
        'details': ['foo', 'bar'],
        'things': ['foo', 'bar'],
    }



# -*- coding: utf-8 -*-

"""Collection of tests around common path and url shorthands."""

from __future__ import unicode_literals
import pytest

from cookiecutter.config import BUILTIN_ABBREVIATIONS
from cookiecutter.repository import expand_abbreviations


@pytest.mark.parametrize(
    ('template', 'abbreviations', 'expected_result'),
    [
        ('foo', {'foo': 'bar'}, 'bar'),
        ('baz', {'foo': 'bar'}, 'baz'),
        ('xx:a', {'xx': '<{0}>'}, '<a>'),
        ('gh:a', {'gh': '<{0}>'}, '<a>'),
        ('xx:a', {'xx': '<>'}, '<>'),
        (
            'gh:pydanny/cookiecutter-django',
            BUILTIN_ABBREVIATIONS,
            'https://github.com/pydanny/cookiecutter-django.git',
        ),
        (
            'gl:pydanny/cookiecutter-django',
            BUILTIN_ABBREVIATIONS,
            'https://gitlab.com/pydanny/cookiecutter-django.git',
        ),
        (
            'bb:pydanny/cookiecutter-django',
            BUILTIN_ABBREVIATIONS,
            'https://bitbucket.org/pydanny/cookiecutter-django',
        ),
    ],
    ids=(
        'Simple expansion',
        'Skip expansion (expansion not an abbreviation)',
        'Expansion prefix',
        'expansion_override_builtin',
        'expansion_prefix_ignores_suffix',
        'Correct expansion for builtin abbreviations (github)',
        'Correct expansion for builtin abbreviations (gitlab)',
        'Correct expansion for builtin abbreviations (bitbucket)',
    ),
)
def test_abbreviation_expansion(template, abbreviations, expected_result):
    """Verify abbreviation unpacking."""
    expanded = expand_abbreviations(template, abbreviations)
    assert expanded == expected_result


def test_abbreviation_expansion_prefix_not_0_in_braces():
    """Verify abbreviation unpacking raises error on incorrect index."""
    with pytest.raises(IndexError):
        expand_abbreviations('xx:a', {'xx': '{1}'})



# -*- coding: utf-8 -*-

"""
cookiecutter.replay.

-------------------
"""

from __future__ import unicode_literals

import json
import os

import six

from cookiecutter.utils import make_sure_path_exists


def get_file_name(replay_dir, template_name):
    """Get the name of file."""
    file_name = '{}.json'.format(template_name)
    return os.path.join(replay_dir, file_name)


def dump(replay_dir, template_name, context, context_key=None):
    """Write json data to file."""
    if not context_key:
        context_key = next(iter(context))

    if not make_sure_path_exists(replay_dir):
        raise IOError('Unable to create replay dir at {}'.format(replay_dir))

    if not isinstance(template_name, six.string_types):
        raise TypeError('Template name is required to be of type str')

    if not isinstance(context, dict):
        raise TypeError('Context is required to be of type dict')

    if context_key not in context:
        raise ValueError('Context is required to contain a cookiecutter key')

    replay_file = get_file_name(replay_dir, template_name)

    with open(replay_file, 'w') as outfile:
        json.dump(context, outfile, indent=2)


def load(replay_dir, template_name, context_key):
    """Read json data from file."""
    if not isinstance(template_name, six.string_types):
        raise TypeError('Template name is required to be of type str')

    replay_file = get_file_name(replay_dir, template_name)

    with open(replay_file, 'r') as infile:
        context = json.load(infile)

    if context_key not in context:
        raise ValueError('Context does not contain the context_key %s' % context_key)

    return context



# -*- coding: utf-8 -*-

"""Functions for finding Cookiecutter templates and other components."""

import logging
import os

logger = logging.getLogger(__name__)


def find_template(repo_dir, context_key='cookiecutter'):
    """Determine which child directory of `repo_dir` is the project template.

    :param repo_dir: Local directory of newly cloned repo.
    :returns project_template: Relative path to project template.
    """
    logger.debug('Searching %s for the project template.', repo_dir)

    repo_dir_contents = os.listdir(repo_dir)

    project_template = None
    for item in repo_dir_contents:
        if context_key in item and '{{' in item and '}}' in item:
            project_template = item
            break

    if project_template:
        project_template = os.path.join(repo_dir, project_template)
        logger.debug('The project template appears to be %s', project_template)
        return project_template
    else:
        return None



# -*- coding: utf-8 -*-

"""Operator plugin that inherits a base class and is made available through `type`."""
from __future__ import unicode_literals
from __future__ import print_function

import sys
import logging

from cookiecutter.operators import BaseOperator
import subprocess

logger = logging.getLogger(__name__)


class CommandOperator(BaseOperator):
    """Operator for PyInquirer type prompts."""

    type = 'command'

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize `command` Hook."""  # noqa
        super(CommandOperator, self).__init__(
            operator_dict=operator_dict, context=context, no_input=no_input
        )
        # Defaulting to run inline
        self.post_gen_operator = (
            self.operator_dict['delay'] if 'delay' in self.operator_dict else False
        )

    def execute(self):
        """Run the operator."""  # noqa
        p = subprocess.Popen(
            self.operator_dict['command'],
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        output, err = p.communicate()

        if err:
            sys.exit(err)

        return output.decode("utf-8")


# TODO: We need a way to spawn a pseudo terminal like session.
#  This should likely be the default for this operator
#
# class CommandInputOperator(BaseOperator):
#     """Operator for PyInquirer type prompts."""
#
#     type = 'command_input'
#
#     def __init__(self, operator_dict, context=None):
#         """Initialize PyInquirer Hook."""  # noqa
#         super(CommandInputOperator, self).__init__(
#             operator_dict=operator_dict, context=context
#         )
#         self.post_gen_operator = True
#
#     def execute(self):
#         """Run the operator."""  # noqa
#         p = subprocess.Popen(
#             self.operator_dict['command'],
#             shell=True,
#             stdout=subprocess.PIPE,
#             stderr=subprocess.PIPE,
#         )
#         output, err = p.communicate()
#
#         if err:
#             sys.exit(err)
#
#         return output.decode("utf-8")



# -*- coding: utf-8 -*-

"""Functions for generating a project from a project template."""
from __future__ import unicode_literals

import fnmatch
import io
import logging
import os
import shutil
from collections import OrderedDict

from binaryornot.check import is_binary
from jinja2 import FileSystemLoader
from jinja2.exceptions import TemplateSyntaxError, UndefinedError

from cookiecutter.environment import StrictEnvironment
from cookiecutter.exceptions import (
    NonTemplatedInputDirException,
    ContextDecodingException,
    FailedHookException,
    OutputDirExistsException,
    UndefinedVariableInTemplate,
)
from cookiecutter.find import find_template
from cookiecutter.hooks import run_hook
from cookiecutter.utils import make_sure_path_exists, work_in, rmtree, read_config_file

# from cookiecutter.repository import valid_context_files

from cookiecutter.operator import post_gen_operator_list

logger = logging.getLogger(__name__)


def is_copy_only_path(path, context, context_key='cookiecutter'):
    """Check whether the given `path` should only be copied and not rendered.

    Returns True if `path` matches a pattern in the given `context` dict,
    otherwise False.

    :param path: A file-system path referring to a file or dir that
        should be rendered or just copied.
    :param context: cookiecutter context.
    """
    try:
        for dont_render in context[context_key]['_copy_without_render']:
            if fnmatch.fnmatch(path, dont_render):
                return True
    except KeyError:
        return False

    return False


def apply_overwrites_to_context(context, overwrite_context):
    """Modify the given context in place based on the overwrite_context."""
    for variable, overwrite in overwrite_context.items():
        if variable not in context:
            # Do not include variables which are not used in the template
            continue

        context_value = context[variable]

        if isinstance(context_value, list):
            # We are dealing with a choice variable
            if overwrite in context_value:
                # This overwrite is actually valid for the given context
                # Let's set it as default (by definition first item in list)
                # see ``cookiecutter.prompt.prompt_choice_for_config``
                context_value.remove(overwrite)
                context_value.insert(0, overwrite)
        else:
            # Simply overwrite the value for this variable
            context[variable] = overwrite


def generate_context(
    # context_file=None, default_context=None, extra_context=None
    context_file='cookiecutter.json',
    default_context=None,
    extra_context=None,
):
    """Generate the context for a Cookiecutter project template.

    Loads the JSON file as a Python object, with key being the JSON filename.

    :param context_file: JSON file containing key/value pairs for populating
        the cookiecutter's variables.
    :param default_context: Dictionary containing config to take into account.
    :param extra_context: Dictionary containing configuration overrides
    """
    context = OrderedDict([])

    try:
        obj = read_config_file(context_file)
    except ValueError as e:
        # JSON decoding error.  Let's throw a new exception that is more
        # friendly for the developer or user.
        full_fpath = os.path.abspath(context_file)
        json_exc_message = str(e)
        our_exc_message = (
            'JSON decoding error while loading "{0}".  Decoding'
            ' error details: "{1}"'.format(full_fpath, json_exc_message)
        )
        raise ContextDecodingException(our_exc_message)

    # Add the Python object to the context dictionary
    file_name = os.path.split(context_file)[1]
    file_stem = file_name.split('.')[0]
    context[file_stem] = obj

    # Overwrite context variable defaults with the default context from the
    # user's global config, if available
    if default_context:
        apply_overwrites_to_context(obj, default_context)
    if extra_context:
        apply_overwrites_to_context(obj, extra_context)

    logger.debug('Context generated is %s', context)
    return context


def generate_file(project_dir, infile, context, env, skip_if_file_exists=False):
    """Render filename of infile as name of outfile, handle infile correctly.

    Dealing with infile appropriately:

        a. If infile is a binary file, copy it over without rendering.
        b. If infile is a text file, render its contents and write the
           rendered infile to outfile.

    Precondition:

        When calling `generate_file()`, the root template dir must be the
        current working directory. Using `utils.work_in()` is the recommended
        way to perform this directory change.

    :param project_dir: Absolute path to the resulting generated project.
    :param infile: Input file to generate the file from. Relative to the root
        template dir.
    :param context: Dict for populating the cookiecutter's variables.
    :param env: Jinja2 template execution environment.
    """
    logger.debug('Processing file %s', infile)

    # Render the path to the output file (not including the root project dir)
    outfile_tmpl = env.from_string(infile)

    outfile = os.path.join(project_dir, outfile_tmpl.render(**context))
    file_name_is_empty = os.path.isdir(outfile)
    if file_name_is_empty:
        logger.debug('The resulting file name is empty: %s', outfile)
        return

    if skip_if_file_exists and os.path.exists(outfile):
        logger.debug('The resulting file already exists: %s', outfile)
        return

    logger.debug('Created file at %s', outfile)

    # Just copy over binary files. Don't render.
    logger.debug("Check %s to see if it's a binary", infile)
    if is_binary(infile):
        logger.debug('Copying binary %s to %s without rendering', infile, outfile)
        shutil.copyfile(infile, outfile)
    else:
        # Force fwd slashes on Windows for get_template
        # This is a by-design Jinja issue
        infile_fwd_slashes = infile.replace(os.path.sep, '/')

        # Render the file
        try:
            tmpl = env.get_template(infile_fwd_slashes)
        except TemplateSyntaxError as exception:
            # Disable translated so that printed exception contains verbose
            # information about syntax error location
            exception.translated = False
            raise
        rendered_file = tmpl.render(**context)

        logger.debug('Writing contents to file %s', outfile)

        with io.open(outfile, 'w', encoding='utf-8') as fh:
            fh.write(rendered_file)

    # Apply file permissions to output file
    shutil.copymode(infile, outfile)


def render_and_create_dir(
    dirname, context, output_dir, environment, overwrite_if_exists=False
):
    """Render name of a directory, create the directory, return its path."""
    name_tmpl = environment.from_string(dirname)
    rendered_dirname = name_tmpl.render(**context)

    dir_to_create = os.path.normpath(os.path.join(output_dir, rendered_dirname))

    logger.debug(
        'Rendered dir %s must exist in output_dir %s', dir_to_create, output_dir
    )

    output_dir_exists = os.path.exists(dir_to_create)

    if output_dir_exists:
        if overwrite_if_exists:
            logger.debug(
                'Output directory %s already exists, overwriting it', dir_to_create
            )
        else:
            msg = 'Error: "{}" directory already exists'.format(dir_to_create)
            raise OutputDirExistsException(msg)
    else:
        make_sure_path_exists(dir_to_create)

    return dir_to_create, not output_dir_exists


def ensure_dir_is_templated(dirname):
    """Ensure that dirname is a templated directory name."""
    if '{{' in dirname and '}}' in dirname:
        return True
    else:
        raise NonTemplatedInputDirException


def _run_hook_from_repo_dir(
    repo_dir, hook_name, project_dir, context, delete_project_on_failure
):
    """Run hook from repo directory, clean project directory if hook fails.

    :param repo_dir: Project template input directory.
    :param hook_name: The hook to execute.
    :param project_dir: The directory to execute the script from.
    :param context: Cookiecutter project context.
    :param delete_project_on_failure: Delete the project directory on hook
        failure?
    """
    with work_in(repo_dir):
        try:
            run_hook(hook_name, project_dir, context)
        except FailedHookException:
            if delete_project_on_failure:
                rmtree(project_dir)
            logger.error(
                "Stopping generation because %s hook "
                "script didn't exit successfully",
                hook_name,
            )
            raise


def generate_files(
    repo_dir,
    context=None,
    output_dir='.',
    overwrite_if_exists=False,
    skip_if_file_exists=False,
    context_key=None,
):
    """Render the templates and saves them to files.

    :param repo_dir: Project template input directory.
    :param context: Dict for populating the template's variables.
    :param output_dir: Where to output the generated project dir into.
    :param overwrite_if_exists: Overwrite the contents of the output directory
        if it exists.
    """
    if not context_key:
        context_key = next(iter(context))

    template_dir = find_template(repo_dir, context_key)
    if template_dir:
        logger.debug('Generating project from %s...', template_dir)
        context = context or OrderedDict([])

        unrendered_dir = os.path.split(template_dir)[1]
        ensure_dir_is_templated(unrendered_dir)
        env = StrictEnvironment(context=context, keep_trailing_newline=True)
        try:
            project_dir, output_directory_created = render_and_create_dir(
                unrendered_dir, context, output_dir, env, overwrite_if_exists
            )
        except UndefinedError as err:
            msg = "Unable to create project directory '{}'".format(unrendered_dir)
            raise UndefinedVariableInTemplate(msg, err, context)

        # We want the Jinja path and the OS paths to match. Consequently, we'll:
        #   + CD to the template folder
        #   + Set Jinja's path to '.'
        #
        #  In order to build our files to the correct folder(s), we'll use an
        # absolute path for the target folder (project_dir)

        project_dir = os.path.abspath(project_dir)
        logger.debug('Project directory is %s', project_dir)

        # if we created the output directory, then it's ok to remove it
        # if rendering fails
        delete_project_on_failure = output_directory_created

        _run_hook_from_repo_dir(
            repo_dir, 'pre_gen_project', project_dir, context, delete_project_on_failure
        )

        with work_in(template_dir):
            env.loader = FileSystemLoader('.')

            for root, dirs, files in os.walk('.'):
                # We must separate the two types of dirs into different lists.
                # The reason is that we don't want ``os.walk`` to go through the
                # unrendered directories, since they will just be copied.
                copy_dirs = []
                render_dirs = []

                for d in dirs:
                    d_ = os.path.normpath(os.path.join(root, d))
                    # We check the full path, because that's how it can be
                    # specified in the ``_copy_without_render`` setting, but
                    # we store just the dir name
                    if is_copy_only_path(d_, context, context_key):
                        copy_dirs.append(d)
                    else:
                        render_dirs.append(d)

                for copy_dir in copy_dirs:
                    indir = os.path.normpath(os.path.join(root, copy_dir))
                    outdir = os.path.normpath(os.path.join(project_dir, indir))
                    logger.debug(
                        'Copying dir %s to %s without rendering', indir, outdir
                    )
                    shutil.copytree(indir, outdir)

                # We mutate ``dirs``, because we only want to go through these dirs
                # recursively
                dirs[:] = render_dirs
                for d in dirs:
                    unrendered_dir = os.path.join(project_dir, root, d)
                    try:
                        render_and_create_dir(
                            unrendered_dir,
                            context,
                            output_dir,
                            env,
                            overwrite_if_exists,
                        )
                    except UndefinedError as err:
                        if delete_project_on_failure:
                            rmtree(project_dir)
                        _dir = os.path.relpath(unrendered_dir, output_dir)
                        msg = "Unable to create directory '{}'".format(_dir)
                        raise UndefinedVariableInTemplate(msg, err, context)

                for f in files:
                    infile = os.path.normpath(os.path.join(root, f))
                    if is_copy_only_path(infile, context, context_key):
                        outfile_tmpl = env.from_string(infile)
                        outfile_rendered = outfile_tmpl.render(**context)
                        outfile = os.path.join(project_dir, outfile_rendered)
                        logger.debug(
                            'Copying file %s to %s without rendering', infile, outfile
                        )
                        shutil.copyfile(infile, outfile)
                        shutil.copymode(infile, outfile)
                        continue
                    try:
                        generate_file(
                            project_dir, infile, context, env, skip_if_file_exists
                        )
                    except UndefinedError as err:
                        if delete_project_on_failure:
                            rmtree(project_dir)
                        msg = "Unable to create file '{}'".format(infile)
                        raise UndefinedVariableInTemplate(msg, err, context)

        _run_hook_from_repo_dir(
            repo_dir,
            'post_gen_project',
            project_dir,
            context,
            delete_project_on_failure,
        )

        for o in post_gen_operator_list:
            o.execute()

        return project_dir
    else:
        for o in post_gen_operator_list:
            o.execute()
        return None



# -*- coding: utf-8 -*-

"""Collection of tests around cookiecutter's replay feature."""

import os
from cookiecutter.main import cookiecutter


def test_replay_dump_template_name(
    monkeypatch, mocker, user_config_data, user_config_file
):
    """Check that replay_dump is called with a valid template_name.

    Template name must not be a relative path.

    Otherwise files such as ``..json`` are created, which are not just cryptic
    but also later mistaken for replay files of other templates if invoked with
    '.' and '--replay'.

    Change the current working directory temporarily to 'tests/fake-repo-tmpl'
    for this test and call cookiecutter with '.' for the target template.
    """
    monkeypatch.chdir(
        os.path.join(os.path.abspath(os.path.dirname(__file__)), 'fake-repo-tmpl')
    )

    mock_replay_dump = mocker.patch('cookiecutter.main.dump')
    mocker.patch('cookiecutter.main.generate_files')

    cookiecutter(
        '.', no_input=True, replay=False, config_file=user_config_file,
    )

    mock_replay_dump.assert_called_once_with(
        user_config_data['replay_dir'], 'fake-repo-tmpl', mocker.ANY, 'cookiecutter'
    )


def test_replay_load_template_name(
    monkeypatch, mocker, user_config_data, user_config_file
):
    """Check that replay_load is called correctly.

    Calls require valid template_name that is not a relative path.

    Change the current working directory temporarily to 'tests/fake-repo-tmpl'
    for this test and call cookiecutter with '.' for the target template.
    """
    monkeypatch.chdir(
        os.path.join(os.path.abspath(os.path.dirname(__file__)), 'fake-repo-tmpl')
    )

    mock_replay_load = mocker.patch('cookiecutter.main.load')
    mocker.patch('cookiecutter.main.generate_files')

    cookiecutter(
        '.', replay=True, config_file=user_config_file,
    )

    mock_replay_load.assert_called_once_with(
        user_config_data['replay_dir'], 'fake-repo-tmpl', 'cookiecutter'
    )



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.prompt` module."""
import os

from cookiecutter.operator import parse_operator

# from cookiecutter.operator import run_operator

base_dir = os.path.dirname(__file__)
# tpl_file = os.path.join(base_dir, "tpl", "things.py.j2")
# output_file = os.path.join(base_dir, "things.py")

context = {
    'cookiecutter': {
        'project_name': "Slartibartfast",
        'details': {
            "type": "jinja",
            # "template_path": tpl_file,
            # "output_path": output_file,
            "template_path": "templates/things.py.j2",
            "output_path": "things.py",
            "file_system_loader": base_dir,
        },
    }
}


# def test_jinja_operator():
#     """Verify simplest functionality."""
#     if os.path.exists('things.py'):
#         os.remove('things.py')
#
#     operator_output, delayed_output = run_operator(
#         context['cookiecutter']['details'], context
#     )
#
#     expected_output = 'things.py'
#     dir_output = os.listdir(base_dir)
#     expected_dir_output = ['templates', 'test_jinja.py', 'things.py', '__pycache__', '__init__.py'] # noqa
#
#     assert operator_output == expected_output
#     assert dir_output == expected_dir_output
#     assert not delayed_output
#     os.remove('things.py')


def test_jinja_parse_operator():
    """Verify simplest functionality."""
    if os.path.exists('things.py'):
        os.remove('things.py')

    cookiecutter_dict = parse_operator(context, 'details', {})
    assert cookiecutter_dict == {'details': "things.py"}
    with open("things.py") as f:
        output = f.read()
    expected_file_output = "project_name"

    assert expected_file_output in output
    os.remove('things.py')



# -*- coding: utf-8 -*-
"""Test helper for pyinquirer functionality."""

#
# from __future__ import print_function
#
# import os
# import sys
# import codecs
# import time
# import select
#
# import pytest
# from ptyprocess import PtyProcess
# import regex
#
# from prompt_toolkit.eventloop.posix import PosixEventLoop
# from prompt_toolkit.input import PipeInput
# from prompt_toolkit.interface import CommandLineInterface
# from prompt_toolkit.output import DummyOutput
#
# from PyInquirer import style_from_dict, Token
# from PyInquirer import prompts
#
#
# # http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named/?in=user-97991 # noqa
# class Bunch:
#     def __init__(self, **kwds):
#         self.__dict__.update(kwds)
#
#         # use Bunch to create group of variables:
#         # cf = Bunch(datum=y, squared=y*y, coord=x)
#
#
# keys = Bunch(
#     DOWN='\x1b[B',
#     UP='\x1b[A',
#     LEFT='\x1b[D',
#     RIGHT='\x1b[C',
#     ENTER='\x0a',  # ControlJ  (Identical to '\n')
#     ESCAPE='\x1b',
#     CONTROLC='\x03',
#     BACK='\x7f')
#
#
# style = style_from_dict({
#     Token.QuestionMark: '#FF9D00 bold',
#     Token.Selected: '#5F819D bold',
#     Token.Instruction: '',  # default
#     Token.Answer: '#5F819D bold',
#     Token.Question: '',
# })
#
#
# def feed_app_with_input(type, message, text, **kwargs):
#     """
#     Create a CommandLineInterface, feed it with the given user input and return
#     the CLI object.
#
#     This returns a (result, CLI) tuple.
#     note: this only works if you import your prompt and then this function!!
#     """
#     # If the given text doesn't end with a newline, the interface won't finish.
#     assert text.endswith('\n')
#
#     application = getattr(prompts, type).question(message, **kwargs)
#
#     loop = PosixEventLoop()
#     try:
#         inp = PipeInput()
#         inp.send_text(text)
#         cli = CommandLineInterface(
#             application=application,
#             eventloop=loop,
#             input=inp,
#             output=DummyOutput())
#         result = cli.run()
#         return result, cli
#     finally:
#         loop.close()
#         inp.close()
#
#
# def remove_ansi_escape_sequences(text):
#     # http://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python # noqa
#     # remove all ansi escape sequences
#     #return regex.sub(r'(\x9b|\x1b\[)[0-?]*[ -\/]*[@-~]|[ ]*\r', '', text)
#     text = regex.sub(r'(\x9b|\x1b\[)[0-?]*[ -\/]*[@-~]', '', text)
#     text = regex.sub(r'[ \r]*\n', '\n', text)  # also clean up the line endings
#     return text
#
# # helper for running sut as subprocess within pty
# # does two things
# # * test app running in pty in subprocess
# # * get test coverage from subprocess
#
# # docu:
# # http://blog.fizyk.net.pl/blog/gathering-tests-coverage-for-subprocesses-in-python.html # noqa
#
#
# PY3 = sys.version_info[0] >= 3
#
# if PY3:
#     basestring = str
#
#
# class SimplePty(PtyProcess):
#     """Simple wrapper around a process running in a pseudoterminal.
#
#     This class exposes a similar interface to :class:`PtyProcess`, but its read
#     methods return unicode, and its :meth:`write` accepts unicode.
#     """
#
#     def __init__(self, pid, fd, encoding='utf-8', codec_errors='strict'):
#         super(SimplePty, self).__init__(pid, fd)
#         self.encoding = encoding
#         self.codec_errors = codec_errors
#         self.decoder = codecs.getincrementaldecoder(encoding)(errors=codec_errors)
#
#     def read(self, size=1024):
#         """Read at most ``size`` bytes from the pty, return them as unicode.
#
#         Can block if there is nothing to read. Raises :exc:`EOFError` if the
#         terminal was closed.
#
#         The size argument still refers to bytes, not unicode code points.
#         """
#         b = super(SimplePty, self).read(size)
#         if not b:
#             return ''
#         if self.skip_cr:
#             b = b.replace(b'\r', b'')
#         #if self.skip_ansi:
#         #    b = remove_ansi_escape_sequences(b)
#         return self.decoder.decode(b, final=False)
#
#     def readline(self):
#         """Read one line from the pseudoterminal, and return it as unicode.
#
#         Can block if there is nothing to read. Raises :exc:`EOFError` if the
#         terminal was closed.
#         note: this is a specialized version that does not have \r\n at the end
#         """
#         # TODO implement a timeout
#         b = super(SimplePty, self).readline().strip()
#         s = self.decoder.decode(b, final=False)
#         if self.skip_ansi:
#             s = remove_ansi_escape_sequences(s)
#         return s
#
#     def write(self, s):
#         """Write the unicode string ``s`` to the pseudoterminal.
#         This intends to make tests a little less verbose.
#
#         Returns the number of bytes written.
#         """
#         if isinstance(s, basestring):
#             b = s.encode(self.encoding)
#         count = super(SimplePty, self).write(b)
#         return count
#
#     def writeline(self, s):
#         """Syntactic sugar to add a '\n' at the end of the .
#
#         Returns the number of bytes written.
#         """
#         if not s.endswith('\n'):
#             s += '\n'
#         return self.write(s)
#
#     @classmethod
#     def spawn(
#             cls, argv, cwd=None, env=None, echo=False, preexec_fn=None,
#             dimensions=(24, 80), skip_cr=True, skip_ansi=True, timeout=1.0):
#         """
#
#         :param argv:
#         :param cwd:
#         :param env:
#         :param echo: default is False so we do not have to deal with the echo
#         :param preexec_fn:
#         :param dimensions:
#         :param skip_cr: skip carriage return '/r' characters when comparing equality
#         :param skip_ansi: skip ansi escape sequences when comparing equality
#         :param timeout: read timeout in seconds
#         :return: subprocess handle
#         """
#         if env is None:
#             env = os.environ
#         inst = super(SimplePty, cls).spawn(argv, cwd, env, echo, preexec_fn,
#                                            dimensions)
#         inst.skip_cr = skip_cr
#         inst.skip_ansi = skip_ansi
#         inst.timeout = timeout  # in seconds
#         return inst
#
#     def expect(self, text):
#         """Read until equals text or timeout."""
#         # inspired by pexpect/pty_spawn and  pexpect/expect.py expect_loop
#         end_time = time.time() + self.timeout
#         buf = ''
#         while (end_time - time.time()) > 0.0:
#             # switch to nonblocking read
#             reads, _, _ = select.select([self.fd], [], [], end_time - time.time())
#             if len(reads) > 0:
#                 try:
#                     buf = remove_ansi_escape_sequences(buf + self.read())
#                 except EOFError:
#                     print('len: %d' % len(buf))
#                     assert buf == text
#                 if buf == text:
#                     return
#                 elif len(buf) >= len(text):
#                     break
#             else:
#                 # do not eat up CPU when waiting for the timeout to expire
#                 time.sleep(self.timeout/10)
#         #print(repr(buf))  # debug ansi code handling
#         assert buf == text
#
#     def expect_regex(self, pattern):
#         """Read until matches pattern or timeout."""
#         # inspired by pexpect/pty_spawn and  pexpect/expect.py expect_loop
#         end_time = time.time() + self.timeout
#         buf = ''
#         prog = regex.compile(pattern)
#         while (end_time - time.time()) > 0.0:
#             # switch to nonblocking read
#             reads, _, _ = select.select([self.fd], [], [], end_time - time.time())
#             if len(reads) > 0:
#                 try:
#                     buf = remove_ansi_escape_sequences(buf + self.read())
#                 except EOFError:
#                     assert prog.match(buf) is not None, \
#                         'output was:\n%s\nexpect regex pattern:\n%s' % (buf, pattern)
#                 if prog.match(buf):
#                     return True
#             else:
#                 # do not eat up CPU when waiting for the timeout to expire
#                 time.sleep(self.timeout/10)
#         assert prog.match(buf) is not None, \
#             'output was:\n%s\nexpect regex pattern:\n%s' % (buf, pattern)
#
#
# def create_example_fixture(example):
#     """Create a pytest fixture to run the example in pty subprocess & cleanup.
#
#     :param example: relative path like 'examples/input.py'
#     :return: pytest fixture
#     """
#     @pytest.fixture
#     def example_app():
#         p = SimplePty.spawn(['python', example])
#         yield p
#         # it takes some time to collect the coverage data
#         # if the main process exits too early the coverage data is not available
#         time.sleep(p.delayafterterminate)
#         try:
#             p.sendintr()  # in case the subprocess was not ended by the test
#         except OSError as e:
#             if e.errno != 5:
#                 raise
#         p.wait()  # without wait() the coverage info never arrives
#     return example_app



# -*- coding: utf-8 -*-

"""
test_cookiecutter_invocation.

Tests to make sure that cookiecutter can be called from the cli without
using the entry point set up for the package.
"""

import os
import subprocess
import sys

import pytest

from cookiecutter import utils


@pytest.fixture
def project_dir():
    """Return test project folder name and remove it after the test."""
    yield 'fake-project-templated'

    if os.path.isdir('fake-project-templated'):
        utils.rmtree('fake-project-templated')


def test_should_raise_error_without_template_arg(monkeypatch, capfd):
    """Verify expected error in command line on invocation without arguments."""
    monkeypatch.setenv('PYTHONPATH', '.')

    with pytest.raises(subprocess.CalledProcessError):
        subprocess.check_call([sys.executable, '-m', 'cookiecutter.cli'])

    _, err = capfd.readouterr()
    exp_message = "Error: Missing argument"
    assert exp_message in err and "TEMPLATE" in err


@pytest.mark.usefixtures('clean_system')
def test_should_invoke_main(monkeypatch, project_dir):
    """Should create a project and exit with 0 code on cli invocation."""
    monkeypatch.setenv('PYTHONPATH', '.')
    test_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..')
    monkeypatch.chdir(test_dir)

    exit_code = subprocess.check_call(
        [sys.executable, '-m', 'cookiecutter.cli', 'tests/fake-repo-tmpl', '--no-input']
    )
    assert exit_code == 0
    assert os.path.isdir(project_dir)


@pytest.mark.usefixtures('clean_system')
def test_should_invoke_main_nuki(monkeypatch, project_dir):
    """Should create a project and exit with 0 code on cli invocation."""
    monkeypatch.setenv('PYTHONPATH', '.')
    test_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..')
    monkeypatch.chdir(test_dir)

    exit_code = subprocess.check_call(
        [
            sys.executable,
            '-m',
            'cookiecutter.cli',
            'tests/fake-repo-tmpl-nuki',
            '--no-input',
        ]
    )

    assert exit_code == 0
    assert os.path.isdir(project_dir)


# @pytest.mark.usefixtures('clean_system')
# def test_should_invoke_main_nuki_nukis(monkeypatch, project_dir):
#     """Should create a project and exit with 0 code on cli invocation."""
#     monkeypatch.setenv('PYTHONPATH', '.')
#     test_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..')
#     monkeypatch.chdir(test_dir)
#
#     exit_code = subprocess.check_call(
#         [
#             sys.executable,
#             '-m',
#             'cookiecutter.cli',
#             'tests/fake-repo-tmpl-nukis',
#             '--no-input',
#         ]
#     )
#
#     assert exit_code == 0
#     assert os.path.isdir(project_dir)



# -*- coding: utf-8 -*-

"""
Main entry point for the `cookiecutter` command.

The code in this module is also a good example of how to use Cookiecutter as a
library rather than a script.
"""

from __future__ import unicode_literals
import logging
import os

from cookiecutter.config import get_user_config
from cookiecutter.generate import generate_context, generate_files
from cookiecutter.exceptions import InvalidModeException
from cookiecutter.prompt import prompt_for_config
from cookiecutter.replay import dump, load
from cookiecutter.repository import determine_repo_dir
from cookiecutter.utils import rmtree

logger = logging.getLogger(__name__)


def cookiecutter(
    template=None,
    checkout=None,
    no_input=False,
    context_file=None,
    context_key=None,
    extra_context=None,
    replay=False,
    overwrite_if_exists=False,
    output_dir='.',
    config_file=None,
    default_config=False,
    password=None,
    directory=None,
    skip_if_file_exists=False,
):
    """
    Run Cookiecutter just as if using it from the command line.

    :param template: A directory containing a project template directory,
        or a URL to a git repository.
    :param checkout: The branch, tag or commit ID to checkout after clone.
    :param no_input: Prompt the user at command line for manual configuration?
    :param extra_context: A dictionary of context that overrides default
        and user configuration.
    :param overwrite_if_exists: Overwrite the contents of output directory
        if it exists
    :param output_dir: Where to output the generated project dir into.
    :param config_file: User configuration file path.
    :param default_config: Use default values rather than a config file.
    :param password: The password to use when extracting the repository.
    :param directory: Relative path to a cookiecutter template in a repository.
    """
    if replay and ((no_input is not False) or (extra_context is not None)):
        err_msg = (
            "You can not use both replay and no_input or extra_context "
            "at the same time."
        )
        raise InvalidModeException(err_msg)

    config_dict = get_user_config(
        config_file=config_file, default_config=default_config,
    )

    repo_dir, context_file, cleanup = determine_repo_dir(
        template=template,
        abbreviations=config_dict['abbreviations'],
        clone_to_dir=config_dict['cookiecutters_dir'],
        checkout=checkout,
        no_input=no_input,
        context_file=context_file,
        password=password,
        directory=directory,
    )

    template_name = os.path.basename(os.path.abspath(repo_dir))

    if replay:
        if not context_key:
            context_key = context_file.split('.')[0]
        context = load(config_dict['replay_dir'], template_name, context_key)
    else:
        context_file_path = os.path.join(repo_dir, context_file)
        logger.debug('context_file is %s', context_file_path)

        context = generate_context(
            context_file=context_file_path,
            default_context=config_dict['default_context'],
            extra_context=extra_context,
        )

        if not context_key:
            context_key = context_file.split('.')[0]
        # prompt the user to manually configure at the command line.pyth
        # except when 'no-input' flag is set
        context[context_key] = prompt_for_config(context, no_input, context_key)

        # include template dir or url in the context dict
        context[context_key]['_template'] = template

        dump(config_dict['replay_dir'], template_name, context, context_key)

    # Create project from local context and project template.
    result = generate_files(
        repo_dir=repo_dir,
        context=context,
        overwrite_if_exists=overwrite_if_exists,
        skip_if_file_exists=skip_if_file_exists,
        output_dir=output_dir,
        context_key=context_key,
    )
    if result:
        logger.debug('Resulting project directory created at %s', result)
    else:
        logger.debug('No project directory was created')

    # Cleanup (if required)
    if cleanup:
        rmtree(repo_dir)

    return context



# -*- coding: utf-8 -*-

"""All exceptions used in the Cookiecutter code base are defined here."""


class CookiecutterException(Exception):
    """
    Base exception class.

    All Cookiecutter-specific exceptions should subclass this class.
    """


class NonTemplatedInputDirException(CookiecutterException):
    """
    Exception for when a project's input dir is not templated.

    The name of the input directory should always contain a string that is
    rendered to something else, so that input_dir != output_dir.
    """


class UnknownTemplateDirException(CookiecutterException):
    """
    Exception for ambiguous project template directory.

    Raised when Cookiecutter cannot determine which directory is the project
    template, e.g. more than one dir appears to be a template dir.
    """

    # unused locally


class MissingProjectDir(CookiecutterException):
    """
    Exception for missing generated project directory.

    Raised during cleanup when remove_repo() can't find a generated project
    directory inside of a repo.
    """

    # unused locally


class ConfigDoesNotExistException(CookiecutterException):
    """
    Exception for missing config file.

    Raised when get_config() is passed a path to a config file, but no file
    is found at that path.
    """


class InvalidConfiguration(CookiecutterException):
    """
    Exception for invalid configuration file.

    Raised if the global configuration file is not valid YAML or is
    badly constructed.
    """


class UnknownRepoType(CookiecutterException):
    """
    Exception for unknown repo types.

    Raised if a repo's type cannot be determined.
    """


class VCSNotInstalled(CookiecutterException):
    """
    Exception when version control is unavailable.

    Raised if the version control system (git or hg) is not installed.
    """


class ContextDecodingException(CookiecutterException):
    """
    Exception for failed JSON decoding.

    Raised when a project's JSON context file can not be decoded.
    """


class OutputDirExistsException(CookiecutterException):
    """
    Exception for existing output directory.

    Raised when the output directory of the project exists already.
    """


class InvalidModeException(CookiecutterException):
    """
    Exception for incompatible modes.

    Raised when cookiecutter is called with both `no_input==True` and
    `replay==True` at the same time.
    """


class FailedHookException(CookiecutterException):
    """
    Exception for hook failures.

    Raised when a hook script fails.
    """


class UndefinedVariableInTemplate(CookiecutterException):
    """
    Exception for out-of-scope variables.

    Raised when a template uses a variable which is not defined in the
    context.
    """

    def __init__(self, message, error, context):
        """Exception for out-of-scope variables."""
        self.message = message
        self.error = error
        self.context = context

    def __str__(self):
        """Text representation of UndefinedVariableInTemplate."""
        return (
            "{self.message}. "
            "Error message: {self.error.message}. "
            "Context: {self.context}"
        ).format(**locals())


class UnknownExtension(CookiecutterException):
    """
    Exception for un-importable extention.

    Raised when an environment is unable to import a required extension.
    """


class RepositoryNotFound(CookiecutterException):
    """
    Exception for missing repo.

    Raised when the specified cookiecutter repository doesn't exist.
    """


class RepositoryCloneFailed(CookiecutterException):
    """
    Exception for un-cloneable repo.

    Raised when a cookiecutter template can't be cloned.
    """


class InvalidZipRepository(CookiecutterException):
    """
    Exception for bad zip repo.

    Raised when the specified cookiecutter repository isn't a valid
    Zip archive.
    """



# -*- coding: utf-8 -*-

"""Cookiecutter repository functions."""

from __future__ import unicode_literals
import os
import re

from cookiecutter.exceptions import RepositoryNotFound
from cookiecutter.vcs import clone
from cookiecutter.zipfile import unzip

REPO_REGEX = re.compile(
    r"""
# something like git:// ssh:// file:// etc.
((((git|hg)\+)?(git|ssh|file|https?):(//)?)
 |                                      # or
 (\w+@[\w\.]+)                          # something like user@...
)
""",
    re.VERBOSE,
)

valid_context_files = [
    'nuki.json',
    'nuki.yaml',
    'nuki.yml',
    'nukikata.json',
    'nukikata.yaml',
    'nukikata.yml',
    'cookiecutter.json',
    'cookiecutter.yaml',
    'cookiecutter.yml',
]


def is_repo_url(value):
    """Return True if value is a repository URL."""
    return bool(REPO_REGEX.match(value))


def is_zip_file(value):
    """Return True if value is a zip file."""
    return value.lower().endswith('.zip')


def expand_abbreviations(template, abbreviations):
    """Expand abbreviations in a template name.

    :param template: The project template name.
    :param abbreviations: Abbreviation definitions.
    """
    if template in abbreviations:
        return abbreviations[template]

    # Split on colon. If there is no colon, rest will be empty
    # and prefix will be the whole template
    prefix, sep, rest = template.partition(':')
    if prefix in abbreviations:
        return abbreviations[prefix].format(rest)

    return template


def repository_has_cookiecutter_json(repo_directory, context_file=None):
    """Determine if `repo_directory` contains a `cookiecutter.json` file.

    :param repo_directory: The candidate repository directory.
    :param context_file:
    :return: True if the `repo_directory` is valid, else False.
    """
    repo_directory_exists = os.path.isdir(repo_directory)
    if context_file:
        context_file_exists = os.path.isfile(
            os.path.join(os.path.abspath(repo_directory), context_file)
        )
        if context_file_exists:
            # The supplied context file exists
            return context_file

    if repo_directory_exists:
        # Check for valid context files as default
        for f in valid_context_files:
            if os.path.isfile(os.path.join(repo_directory, f)):
                return f
    else:
        return False


def determine_repo_dir(
    template,
    abbreviations,
    clone_to_dir,
    checkout,
    no_input,
    context_file=None,
    password=None,
    directory=None,
):
    """
    Locate the repository directory from a template reference.

    Applies repository abbreviations to the template reference.
    If the template refers to a repository URL, clone it.
    If the template is a path to a local repository, use it.

    :param template: A directory containing a project template directory,
        or a URL to a git repository.
    :param abbreviations: A dictionary of repository abbreviation
        definitions.
    :param clone_to_dir: The directory to clone the repository into.
    :param checkout: The branch, tag or commit ID to checkout after clone.
    :param no_input: Prompt the user at command line for manual configuration?
    :param password: The password to use when extracting the repository.
    :param directory: Directory within repo where cookiecutter.json lives.
    :return: A tuple containing the cookiecutter template directory, and
        a boolean descriving whether that directory should be cleaned up
        after the template has been instantiated.
    :raises: `RepositoryNotFound` if a repository directory could not be found.
    """
    template = expand_abbreviations(template, abbreviations)

    if is_zip_file(template):
        unzipped_dir = unzip(
            zip_uri=template,
            is_url=is_repo_url(template),
            clone_to_dir=clone_to_dir,
            no_input=no_input,
            password=password,
        )
        repository_candidates = [unzipped_dir]
        cleanup = True
    elif is_repo_url(template):
        cloned_repo = clone(
            repo_url=template,
            checkout=checkout,
            clone_to_dir=clone_to_dir,
            no_input=no_input,
        )
        repository_candidates = [cloned_repo]
        cleanup = False
    else:
        repository_candidates = [template, os.path.join(clone_to_dir, template)]
        cleanup = False

    if directory:
        repository_candidates = [
            os.path.join(s, directory) for s in repository_candidates
        ]

    for repo_candidate in repository_candidates:
        context_file = repository_has_cookiecutter_json(repo_candidate, context_file)
        if not context_file:
            # Means that no valid context file has been found or provided
            continue
        else:
            return repo_candidate, context_file, cleanup

    raise RepositoryNotFound(
        'A valid repository for "{}" could not be found in the following '
        'locations:\n{}'.format(template, '\n'.join(repository_candidates))
    )



# -*- coding: utf-8 -*-

"""Verify Jinja2 filters/extensions are available from pre-gen/post-gen hooks."""

import io
import os
from _collections import OrderedDict

import freezegun
import pytest

from cookiecutter.main import cookiecutter


@pytest.fixture(autouse=True)
def freeze():
    """Fixture. Make time stating during all tests in this file."""
    freezer = freezegun.freeze_time("2015-12-09 23:33:01")
    freezer.start()
    yield
    freezer.stop()


def test_jinja2_time_extension(monkeypatch, tmpdir):
    """Verify Jinja2 time extension work correctly."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    context = cookiecutter(
        'test-extensions/default/', no_input=True, output_dir=str(tmpdir)
    )

    assert type(context) == OrderedDict

    changelog_file = os.path.join(tmpdir, os.listdir(tmpdir)[0], 'HISTORY.rst')
    assert os.path.isfile(changelog_file)

    with io.open(changelog_file, 'r', encoding='utf-8') as f:
        changelog_lines = f.readlines()

    expected_lines = [
        'History\n',
        '-------\n',
        '\n',
        '0.1.0 (2015-12-09)\n',
        '---------------------\n',
        '\n',
        'First release on PyPI.\n',
    ]
    assert expected_lines == changelog_lines


def test_jinja2_slugify_extension(monkeypatch, tmpdir):
    """Verify Jinja2 slugify extension work correctly."""
    monkeypatch.chdir(os.path.abspath(os.path.dirname(__file__)))

    context = cookiecutter(
        'test-extensions/default/', no_input=True, output_dir=str(tmpdir)
    )

    assert os.listdir(tmpdir)[0] == "it-s-slugified-foobar"
    assert context == OrderedDict(
        [
            (
                'cookiecutter',
                OrderedDict(
                    [
                        ('project_slug', 'it-s-slugified-foobar'),
                        ('year', '2015'),
                        ('_template', 'test-extensions/default/'),
                    ]
                ),
            )
        ]
    )



# -*- coding: utf-8 -*-

"""Jinja2 extensions."""

import json
import string

try:
    # Python 3.6 and above
    from secrets import choice
except ImportError:
    from random import choice

from jinja2.ext import Extension
from slugify import slugify as pyslugify


class JsonifyExtension(Extension):
    """Jinja2 extension to convert a Python object to JSON."""

    def __init__(self, environment):
        """Initialize the extension with the given environment."""
        super(JsonifyExtension, self).__init__(environment)

        def jsonify(obj):
            return json.dumps(obj, sort_keys=True, indent=4)

        environment.filters['jsonify'] = jsonify


class RandomStringExtension(Extension):
    """Jinja2 extension to create a random string."""

    def __init__(self, environment):
        """Jinja2 Extension Constructor."""
        super(RandomStringExtension, self).__init__(environment)

        def random_ascii_string(length, punctuation=False):
            if punctuation:
                corpus = "".join((string.ascii_letters, string.punctuation))
            else:
                corpus = string.ascii_letters
            return "".join(choice(corpus) for _ in range(length))

        environment.globals.update(random_ascii_string=random_ascii_string)


class SlugifyExtension(Extension):
    """Jinja2 Extension to slugify string."""

    def __init__(self, environment):
        """Jinja2 Extension constructor."""
        super(SlugifyExtension, self).__init__(environment)

        def slugify(value, **kwargs):
            """Slugifies the value."""
            return pyslugify(value, **kwargs)

        environment.filters['slugify'] = slugify



# -*- coding: utf-8 -*-

"""Helper functions for working with version control systems."""

from __future__ import unicode_literals
import logging
import os
import subprocess

try:
    from shutil import which
except ImportError:
    from whichcraft import which

from cookiecutter.exceptions import (
    RepositoryNotFound,
    RepositoryCloneFailed,
    UnknownRepoType,
    VCSNotInstalled,
)
from cookiecutter.utils import make_sure_path_exists, prompt_and_delete

logger = logging.getLogger(__name__)


BRANCH_ERRORS = [
    'error: pathspec',
    'unknown revision',
]


def identify_repo(repo_url):
    """Determine if `repo_url` should be treated as a URL to a git or hg repo.

    Repos can be identified by prepending "hg+" or "git+" to the repo URL.

    :param repo_url: Repo URL of unknown type.
    :returns: ('git', repo_url), ('hg', repo_url), or None.
    """
    repo_url_values = repo_url.split('+')
    if len(repo_url_values) == 2:
        repo_type = repo_url_values[0]
        if repo_type in ["git", "hg"]:
            return repo_type, repo_url_values[1]
        else:
            raise UnknownRepoType
    else:
        if 'git' in repo_url:
            return 'git', repo_url
        elif 'bitbucket' in repo_url:
            return 'hg', repo_url
        else:
            raise UnknownRepoType


def is_vcs_installed(repo_type):
    """
    Check if the version control system for a repo type is installed.

    :param repo_type:
    """
    return bool(which(repo_type))


def clone(repo_url, checkout=None, clone_to_dir='.', no_input=False):
    """Clone a repo to the current directory.

    :param repo_url: Repo URL of unknown type.
    :param checkout: The branch, tag or commit ID to checkout after clone.
    :param clone_to_dir: The directory to clone to.
                         Defaults to the current directory.
    :param no_input: Suppress all user prompts when calling via API.
    :returns: str with path to the new directory of the repository.
    """
    # Ensure that clone_to_dir exists
    clone_to_dir = os.path.expanduser(clone_to_dir)
    make_sure_path_exists(clone_to_dir)

    # identify the repo_type
    repo_type, repo_url = identify_repo(repo_url)

    # check that the appropriate VCS for the repo_type is installed
    if not is_vcs_installed(repo_type):
        msg = "'{0}' is not installed.".format(repo_type)
        raise VCSNotInstalled(msg)

    repo_url = repo_url.rstrip('/')
    repo_name = os.path.split(repo_url)[1]
    if repo_type == 'git':
        repo_name = repo_name.split(':')[-1].rsplit('.git')[0]
        repo_dir = os.path.normpath(os.path.join(clone_to_dir, repo_name))
    elif repo_type == 'hg':
        repo_dir = os.path.normpath(os.path.join(clone_to_dir, repo_name))
    logger.debug('repo_dir is {0}'.format(repo_dir))

    if os.path.isdir(repo_dir):
        clone = prompt_and_delete(repo_dir, no_input=no_input)
    else:
        clone = True

    if clone:
        try:
            subprocess.check_output(
                [repo_type, 'clone', repo_url],
                cwd=clone_to_dir,
                stderr=subprocess.STDOUT,
            )
            if checkout is not None:
                subprocess.check_output(
                    [repo_type, 'checkout', checkout],
                    cwd=repo_dir,
                    stderr=subprocess.STDOUT,
                )
        except subprocess.CalledProcessError as clone_error:
            output = clone_error.output.decode('utf-8')
            if 'not found' in output.lower():
                raise RepositoryNotFound(
                    'The repository {} could not be found, '
                    'have you made a typo?'.format(repo_url)
                )
            if any(error in output for error in BRANCH_ERRORS):
                raise RepositoryCloneFailed(
                    'The {} branch of repository {} could not found, '
                    'have you made a typo?'.format(checkout, repo_url)
                )
            raise

    return repo_dir



# -*- coding: utf-8 -*-

"""Tests dict input objects for `cookiecutter.prompt` module."""

import yaml
import os
import shutil

from cookiecutter import prompt

FIXTURES = ['checkbox_bad.yaml']


def checkbox():
    """Test mo stuff."""
    fixture = 'checkbox_bad.yaml'
    with open(fixture, 'r') as f:
        context = yaml.load(f, Loader=yaml.FullLoader)

    if os.path.exists('./best_eva'):
        shutil.rmtree('best_eva')

    out_dict = prompt.prompt_for_config(context)
    return out_dict


if __name__ == '__main__':
    checkbox()

# class TestPromptHookDicts(object):
#     """Class to test all the dict input objects."""
#
#     @pytest.mark.parametrize("input_dicts,invalid", FIXTURES)
#     def test_hook_checkbox(self, monkeypatch):
#         """Verify `prompt_for_config` call `read_user_variable` on dict request."""
#
#         monkeypatch.setattr(
#             'cookiecutter.prompt.read_user_dict',
#             lambda var, default: {"key": "value", "integer": 37},
#         )
#         context = {'cookiecutter': {'details': {}}}
#
#         cookiecutter_dict = prompt.prompt_for_config(context)
#         assert cookiecutter_dict == {'details': {'key': u'value', 'integer': 37}}



# -*- coding: utf-8 -*-

"""Operator plugin that inherits a base class and is made available through `type`."""
from __future__ import unicode_literals
from __future__ import print_function

import logging
import yaml
from cookiecutter.operators import BaseOperator

logger = logging.getLogger(__name__)


class YamlOperator(BaseOperator):
    """Operator for yaml."""

    type = 'yaml'

    def __init__(self, operator_dict, context=None, no_input=False):
        """Initialize yaml operator."""  # noqa
        super(YamlOperator, self).__init__(
            operator_dict=operator_dict, context=context, no_input=no_input
        )
        # Defaulting to run inline
        self.post_gen_operator = (
            self.operator_dict['delay'] if 'delay' in self.operator_dict else False
        )

    def execute(self):
        """Run the operator."""  # noqa
        if 'contents' in self.operator_dict:
            with open(self.operator_dict['path'], 'w') as f:
                yaml.dump(self.operator_dict['contents'], f)

        else:
            with open(self.operator_dict['path'], 'r') as f:
                return yaml.safe_load(f)



# -*- coding: utf-8 -*-

"""Functions for discovering and executing various cookiecutter hooks."""

import errno
import io
import logging
import os
import subprocess
import sys
import tempfile

from cookiecutter import utils
from cookiecutter.environment import StrictEnvironment
from cookiecutter.exceptions import FailedHookException


logger = logging.getLogger(__name__)

_HOOKS = [
    'pre_gen_project',
    'post_gen_project',
]
EXIT_SUCCESS = 0


def valid_hook(hook_file, hook_name):
    """Determine if a hook file is valid.

    :param hook_file: The hook file to consider for validity
    :param hook_name: The hook to find
    :return: The hook file validity
    """
    filename = os.path.basename(hook_file)
    basename = os.path.splitext(filename)[0]

    matching_hook = basename == hook_name
    supported_hook = basename in _HOOKS
    backup_file = filename.endswith('~')

    return matching_hook and supported_hook and not backup_file


def find_hook(hook_name, hooks_dir='hooks'):
    """Return a dict of all hook scripts provided.

    Must be called with the project template as the current working directory.
    Dict's key will be the hook/script's name, without extension, while values
    will be the absolute path to the script. Missing scripts will not be
    included in the returned dict.

    :param hook_name: The hook to find
    :param hooks_dir: The hook directory in the template
    :return: The absolute path to the hook script or None
    """
    logger.debug('hooks_dir is %s', os.path.abspath(hooks_dir))

    if not os.path.isdir(hooks_dir):
        logger.debug('No hooks/dir in template_dir')
        return None

    for hook_file in os.listdir(hooks_dir):
        if valid_hook(hook_file, hook_name):
            return os.path.abspath(os.path.join(hooks_dir, hook_file))

    return None


def run_script(script_path, cwd='.'):
    """Execute a script from a working directory.

    :param script_path: Absolute path to the script to run.
    :param cwd: The directory to run the script from.
    """
    run_thru_shell = sys.platform.startswith('win')
    if script_path.endswith('.py'):
        script_command = [sys.executable, script_path]
    else:
        script_command = [script_path]

    utils.make_executable(script_path)

    try:
        proc = subprocess.Popen(script_command, shell=run_thru_shell, cwd=cwd)
        exit_status = proc.wait()
        if exit_status != EXIT_SUCCESS:
            raise FailedHookException(
                'Hook script failed (exit status: {})'.format(exit_status)
            )
    except OSError as os_error:
        if os_error.errno == errno.ENOEXEC:
            raise FailedHookException(
                'Hook script failed, might be an ' 'empty file or missing a shebang'
            )
        raise FailedHookException('Hook script failed (error: {})'.format(os_error))


def run_script_with_context(script_path, cwd, context):
    """Execute a script after rendering it with Jinja.

    :param script_path: Absolute path to the script to run.
    :param cwd: The directory to run the script from.
    :param context: Cookiecutter project template context.
    """
    _, extension = os.path.splitext(script_path)

    with io.open(script_path, 'r', encoding='utf-8') as file:
        contents = file.read()

    with tempfile.NamedTemporaryFile(delete=False, mode='wb', suffix=extension) as temp:
        env = StrictEnvironment(context=context, keep_trailing_newline=True)
        template = env.from_string(contents)
        output = template.render(**context)
        temp.write(output.encode('utf-8'))

    run_script(temp.name, cwd)


def run_hook(hook_name, project_dir, context):
    """
    Try to find and execute a hook from the specified project directory.

    :param hook_name: The hook to execute.
    :param project_dir: The directory to execute the script from.
    :param context: Cookiecutter project context.
    """
    script = find_hook(hook_name)
    if script is None:
        logger.debug('No %s hook found', hook_name)
        return
    logger.debug('Running hook %s', hook_name)
    run_script_with_context(script, project_dir, context)
